// Copyright (C) 2015 by Bitvise Limited
//
// This header file defines the interface between the Bitvise SSH Server file
// transfer subsystems, and pluggable filesystem providers that can be used
// for access to custom filesystem types.
//
// All callbacks (HandleSfsInit, HandleSfsRequest, HandleSfsDestroy) are
// called from a fiber with a very small stack. Do not allocate memory on the
// stack, and do not use very deep recursion.
// 
// For performance reasons, it is better to implement HandleSfsRequest as
// non-blocking. This involves waiting for an event to happen using
// AddWaitObject, AddTimer, Wait.
//
// HandleSfsRequest can exit without calling SendResponse. In this case,
// HandleSfsRequest should call AcquireAsyncContext and pass it to a thread
// which will eventually call SendResponse and ReleaseAsyncContext.
//
// After a HandleSfsRequest context has called SendResponse, this context may
// no longer be used for logging (using SendLog or SendLogByHandle).

#pragma once

enum { SFS_VERSION = 1 };



// SFS enumerations

struct SFS_AceFlag { enum {
    FileInherit				= 0x00000001,
    DirectoryInherit		= 0x00000002,
    NoPropagateInherit		= 0x00000004,
    InheritOnly				= 0x00000008,
    SuccessfulAccess		= 0x00000010,
    FailedAccess			= 0x00000020,
    IdentifierGroup			= 0x00000040,
	Inherited				= 0x01000000, }; };

struct SFS_AceMask { enum {
    Read					= 0x00000001,
    Write					= 0x00000002,
    Append					= 0x00000004,
    ReadNamedAttrs			= 0x00000008,
    WriteNamedAttrs			= 0x00000010,
    Execute					= 0x00000020,
    DeleteChild				= 0x00000040,
    ReadAttrs				= 0x00000080,
    WriteAttrs				= 0x00000100,
    Delete					= 0x00010000,
    ReadAcl					= 0x00020000,
    WriteAcl				= 0x00040000,
    WriteOwner				= 0x00080000,
    Synchronize				= 0x00100000,
	AccessAuditAlarm		= 0x01000000, }; };

struct SFS_AceType { enum {
	AccessAllowed			= 0,
	AccessDenied			= 1,
	SystemAudit				= 2,
	SystemAlarm				= 3, }; };

struct SFS_AclCap { enum {
    Allow					= 0x0001,
    Deny					= 0x0002,
    Audit					= 0x0004,
    Alarm					= 0x0008,
    InheritAccess			= 0x0010,
	InheritAuditAlarm		= 0x0020, }; };

struct SFS_AclFlag { enum {
    ControlIncluded			= 0x0001,
    ControlPresent			= 0x0002,
    ControlInherited		= 0x0004,
    AuditAlarmIncluded		= 0x0010,
	AuditAlarmInherited		= 0x0020, }; };

struct SFS_Attr { enum {
    Size					= 0x000001,
    Permissions				= 0x000004,
    AccessTime				= 0x000008,
    CreateTime				= 0x000010,
    ModifyTime				= 0x000020,
    Acl						= 0x000040,
    OwnerGroup				= 0x000080,
    Subseconds				= 0x000100,
    Bits					= 0x000200,
    AllocSize				= 0x000400,
    TextHint				= 0x000800,
    MimeType				= 0x001000,
    LinkCount				= 0x002000,
	CTime					= 0x008000, }; };

struct SFS_Type { enum {
    Regular					= 1,
    Directory				= 2,
    Symlink					= 3,
    Special					= 4,
    Unknown					= 5,
    Socket					= 6,
    CharDevice				= 7,
    BlockDevice				= 8,
	Fifo					= 9, }; };

struct SFS_AttrBit { enum {
    ReadOnly				= 0x0001,
    System					= 0x0002,
    Hidden					= 0x0004,
    CaseInsensitive			= 0x0008,
    Archive					= 0x0010,
    Encrypted				= 0x0020,
    Compressed				= 0x0040,
    Sparse					= 0x0080,
    AppendOnly				= 0x0100,
    Immutable				= 0x0200,
	Sync					= 0x0400 }; };

struct SFS_TextHint { enum {
    KnownText				= 0,
    GuessedText				= 1,
    KnownBinary				= 2,
	GuessedBinary			= 3, }; };

struct SFS_OpenFlag { enum {
	AccessDispositionMask	= 0x000007,
		CreateNew				= 0,	// AccessDisposition
		CreateTruncate			= 1,	// AccessDisposition	
		OpenExisting			= 2,	// AccessDisposition
		OpenOrCreate			= 3,	// AccessDisposition
		TruncateExisting		= 4,	// AccessDisposition
    Append					= 0x000008,
    AppendAtomic			= 0x000010,
    TextMode				= 0x000020,
    BlockRead				= 0x000040,
    BlockWrite				= 0x000080,
    BlockDelete				= 0x000100,
    BlockAdvisory			= 0x000200,
    NoFollow				= 0x000400,
    DeleteOnClose			= 0x000800,
    AccessAuditAlarmInfo	= 0x001000,
    AccessBackup			= 0x002000,
    BackupStream			= 0x004000,
	OverrideOwner			= 0x008000, }; };

struct SFS_Permission { enum {
	WorldExecute			= 0x0001,
	WorldWrite				= 0x0002,
	WorldRead				= 0x0004,
	GroupExecute			= 0x0008,
	GroupWrite				= 0x0010,
	GroupRead				= 0x0020,
	OwnerExecute			= 0x0040,
	OwnerWrite				= 0x0080,
	OwnerRead				= 0x0100,
	SaveTextImage			= 0x0200,
	SetGuid					= 0x0400,
	SetUid					= 0x0800, }; };

struct SFS_ControlByte { enum {
	NoCheck					= 1,
	StatIf					= 2,
	StatAlways				= 3, }; };

struct SFS_RenameFlag { enum {
    Overwrite				= 0x01,
    Atomic					= 0x02,
	Native					= 0x04, }; };

struct SFS_RequestType { enum {
    Open					= 3,
    Close					= 4,
    Read					= 5,
    Write					= 6,
    FStat					= 8,
    SetStat					= 9,
    FSetStat				= 10,
    OpenDir					= 11,
    ReadDir					= 12,
    Remove					= 13,
    MkDir					= 14,
    RmDir					= 15,
    RealPath				= 16,
    Stat					= 17,
    Rename					= 18, // *
    ReadLink				= 19,
    Link					= 21,
    Block					= 22,
    Unblock					= 23,
    TextSeek				= 1001,
    CheckFileHandle			= 2001,
    CheckFileName			= 2002,
    SpaceAvail				= 2003,
	FileCopy				= 2005, // *
	DataCopy				= 2006, // *
	NativePath				= 3001, }; };

// * Rename (for files only), FileCopy, and DataCopy will be emulated with the
//   help of NativePath, Open, Read, Write, FStat, FSetStat, Close, and Remove
//   requests if the original request fails with OpUnsupported.

struct SFS_ResponseType { enum {
    Status					= 101,
    Handle					= 102,
    Data					= 103,
    Names					= 104,
    Attrs					= 105,
    Name					= 1101,
    CheckFileReply			= 2101,
	SpaceAvailReply			= 2102, }; };

struct SFS_StatusCode { enum {
    Ok						= 0,
    Eof						= 1,
    NoSuchFile				= 2,
    PermissionDenied		= 3,
    Failure					= 4,
    BadMessage				= 5,
    NoConnection			= 6,
    ConnectionLost			= 7,
    OpUnsupported			= 8,
    InvalidHandle			= 9,
    NoSuchPath				= 10,
    FileAlreadyExists		= 11,
    WriteProtect			= 12,
    NoMedia					= 13,
    NoSpaceOnFilesystem		= 14,
    QuotaExceeded			= 15,
    UnknownPrincipal		= 16,
    LockConflict			= 17,
    DirNotEmpty				= 18,
    NotADirectory			= 19,
    InvalidFilename			= 20,
    LinkLoop				= 21,
    CannotDelete			= 22,
    InvalidParameter		= 23,
    FileIsADirectory		= 24,
    ByteRangeLockConflict	= 25,
    ByteRangeLockRefused	= 26,
    DeletePending			= 27,
    FileCorrupt				= 28,
    OwnerInvalid			= 29,
    GroupInvalid			= 30,
	NoMatchingByteRangeLock	= 31, }; };


struct SFS_LogStatus { enum {
    Success					= 0,
	Failure					= 200,
    Progress				= 600, }; };

struct SFS_LogAction { enum {
	Default					= 00000,
	Init					= 11000,
	Destroy					= 12000,
	Open					= 20000,
	Close					= 21000,
	Read					= 22000,
	Write					= 23000,
	Stat					= 24000,
	SetStat					= 25000,
	OpenDir					= 26000,
	ReadDir					= 27000,
	Remove					= 28000,
	MkDir					= 29000,
	RmDir					= 30000,
	RealPath				= 31000,
	Rename					= 32000,
	ReadLink				= 33000,
	Link					= 34000,
	Block					= 35000,
	Unblock					= 36000,
	TextSeek				= 37000,
	CheckFile				= 38000,
	SpaceAvail				= 39000,
	FileCopy				= 41000,
	DataCopy				= 42000,
	NativePath				= 43000, }; };
	
struct SFS_LogCode { enum {
	CloseSuccess			= SFS_LogAction::Close + SFS_LogStatus::Success,
	TransferFileDone		= 90000,
	ListDirDone				= 91000, }; };


struct SFS_LogItemType { enum {
    Grp						= 0,
	Num						= 2,
	Bln						= 4,
    Str						= 5, }; };

struct SFS_LogItemFlags { enum {
    MultiLine				= 4,
    NewLevel				= 8,
	Hex						= 16, }; };


// SFS common structures
	
struct SFS_Ace
{
	unsigned int m_type;
    unsigned int m_flags;	// SFS_AceFlag
    unsigned int m_mask;	// SFS_AceMask
    wchar_t const* m_who;
};
 
struct SFS_Acl
{
	unsigned int m_flags;
	unsigned int m_aceCount;
	SFS_Ace const* m_aceArray;
};
 
struct SFS_Attrs
{
	unsigned int m_validAttrFlags;	// SFS_Attr
	unsigned int m_type;			// SFS_Type
	unsigned __int64 m_size;
	unsigned __int64 m_allocSize;
	wchar_t const* m_owner;
	wchar_t const* m_group;
	unsigned int m_permissions;
	unsigned __int64 m_accessTime;
	unsigned int m_accessTimeNs;
	unsigned __int64 m_createTime;
	unsigned int m_createTimeNs;
	unsigned __int64 m_modifyTime;
	unsigned int m_modifyTimeNs;
	unsigned __int64 m_cTime;
	unsigned int m_cTimeNs;
	SFS_Acl m_acl;
	unsigned int m_attrBits;		// SFS_AttrBit
	unsigned int m_attrBitsValid;
	unsigned char m_textHint;		// SFS_TextHint
	wchar_t const* m_mimeType;
	unsigned int m_linkCount;
};



// SFS initialization structure

struct SFS_Init
{
	wchar_t const* m_name;
	wchar_t const* m_value;
	unsigned int m_nestedCount;
	SFS_Init const* m_nestedArray;
};



// SFS request structures

struct SFS_Open	// Valid responses: SFS_Status, SFS_Handle
{
	wchar_t const* m_fileName;
	unsigned int m_desiredAccess;	// SFS_AceMask
	unsigned int m_flags;			// SFS_OpenFlag
	SFS_Attrs m_attrs;
};

struct SFS_Close // Valid response: SFS_Status
{
	unsigned __int64 m_handle;
};

struct SFS_Read // Valid responses: SFS_Status, SFS_Data
{
	unsigned __int64 m_handle;
	unsigned __int64 m_offset;
	unsigned int m_length;			// Reading chunks up to 5 MB must be supported
};

struct SFS_Write // Valid response: SFS_Status
{
	unsigned __int64 m_handle;
	unsigned __int64 m_offset;
    unsigned int m_dataSize;
	unsigned char const* m_dataPtr;	// m_dataPtr[0-m_dataSize)
};

struct SFS_FStat // Valid responses: SFS_Status, SFS_Attrs
{
	unsigned __int64 m_handle;
	unsigned int m_flags;	// SFS_Attr
};

struct SFS_SetStat // Valid response: SFS_Status
{
	wchar_t const* m_path;
    SFS_Attrs m_attrs;
};

struct SFS_FSetStat // Valid response: SFS_Status
{
	unsigned __int64 m_handle;
	SFS_Attrs m_attrs;
};

struct SFS_OpenDir // Valid responses: SFS_Status, SFS_Handle
{
	 wchar_t const* m_path;
};

struct SFS_ReadDir // Valid responses: SFS_Status, SFS_Names
{
	 unsigned __int64 m_handle;
};

struct SFS_Remove // Valid response: SFS_Status
{
	wchar_t const* m_fileName;
};

struct SFS_MkDir // Valid response: SFS_Status
{
	wchar_t const* m_path;
	SFS_Attrs m_attrs;
};
 
struct SFS_RmDir // Valid response: SFS_Status
{
	wchar_t const* m_path;
};

struct SFS_RealPath // Valid responses: SFS_Status, SFS_Name
{
	wchar_t const* m_path;
	unsigned char m_controlByte;	// SFS_ControlByte
};

struct SFS_Stat // Valid responses: SFS_Status, SFS_Attrs
{
	wchar_t const* m_path;
	unsigned int m_flags;	// SFS_Attr
	bool m_followSymlink;	// Set false for LStat.
};

struct SFS_Rename // Valid response: SFS_Status
{
	wchar_t const* m_oldPath;
	wchar_t const* m_newPath;
	unsigned int m_flags; // SFS_RenameFlag
};

struct SFS_ReadLink // Valid responses: SFS_Status, SFS_Name
{
	wchar_t const* m_path;
};

struct SFS_Link // Valid response: SFS_Status
{
       wchar_t const* m_newLinkPath;
       wchar_t const* m_existingPath;
	   bool m_symlink;
};

struct SFS_Block // Valid response: SFS_Status
{
	unsigned __int64 m_handle;
	unsigned __int64 m_offset;
	unsigned __int64 m_length;
	unsigned int m_flags;	// SFS_OpenFlag::Block*
};

struct SFS_Unblock // Valid response: SFS_Status
{
	unsigned __int64 m_handle;
	unsigned __int64 m_offset;
	unsigned __int64 m_length;
};

struct SFS_TextSeek // Valid response: SFS_Status
{
	unsigned __int64 m_handle;
	unsigned __int64 m_lineNumber;
};

struct SFS_CheckFileHandle // Valid responses: SFS_Status, SFS_CheckFileReply
{
	unsigned __int64 m_handle;
	wchar_t const* m_hashAlgList;
	unsigned __int64 m_startOffset;
	unsigned __int64 m_length;
	unsigned int m_blockSize;
};
 
struct SFS_CheckFileName // Valid responses: SFS_Status, SFS_CheckFileReply
{
	wchar_t const* m_fileName;
	wchar_t const* m_hashAlgList;
	unsigned __int64 m_startOffset;
	unsigned __int64 m_length;
	unsigned int m_blockSize;
};

struct SFS_SpaceAvail // Valid responses: SFS_Status, SFS_SpaceAvailReply
{
	wchar_t const* m_path;
};

struct SFS_FileCopy // Valid response: SFS_Status
{
	wchar_t const* m_srcFileName;
	wchar_t const* m_dstFileName;
	bool m_overwrite; // Overwrite destination file if exists.
};

struct SFS_DataCopy // Valid response: SFS_Status
{
	unsigned __int64 m_readHandle;
	unsigned __int64 m_readOffset;
    unsigned __int64 m_readLength; // Read until EOF if 0.
    unsigned __int64 m_writeHandle;
    unsigned __int64 m_writeOffset;
};

struct SFS_NativePath // Valid responses: SFS_Status, SFS_Name
{
	bool m_fromHandle;
	unsigned __int64 m_handle;	// Valid if m_fromHandle == true
	wchar_t const* m_path;		// Valid if m_fromHandle == false
};



struct SFS_Request
{
	unsigned int m_type;	// SFS_RequestType
	void const* m_content;
	bool m_internal;		// Ignore transfer and list statistics if true.
};



// SFS response structures

struct SFS_Status
{
	unsigned int m_statusCode;		// SFS_StatusCode
	wchar_t const* m_errorMessage;	// May be null
	wchar_t const* m_languageTag;	// May be null
};

struct SFS_Handle
{
	unsigned __int64 m_handle;
	bool m_createdNewFile;
};
 
struct SFS_Data
{
	unsigned int m_dataSize;
	unsigned char const* m_dataPtr;
	bool m_endOfFile;
};

struct SFS_Name
{
	wchar_t const* m_fileName;
	SFS_Attrs m_attrs;
};

struct SFS_Names
{
	unsigned int m_nameCount;
	SFS_Name const* m_nameArray;	// m_nameArray[0-m_nameCount)
	bool m_endOfList;
};

struct SFS_CheckFileReply
{
	wchar_t const* m_hashAlgUsed;
	unsigned int m_hashDataSize;
	unsigned char const* m_hashDataPtr;	// m_hashDataPtr[0-m_hashDataSize)
};
  
struct SFS_SpaceAvailReply
{
	unsigned __int64 m_bytesOnDevice;
	unsigned __int64 m_unusedBytesOnDevice;
	unsigned __int64 m_bytesAvailableToUser;
	unsigned __int64 m_unusedBytesAvailableToUser;
	unsigned int m_bytesPerAllocationUnit;
};
  
struct SFS_Response
{
	unsigned int m_type;	// SFS_ResponseType
	void const* m_content;			
};



// SFS logging

struct SFS_LogItem
{
	unsigned char m_type;				// SFS_LogItemType
	unsigned char m_flags;				// SFS_LogItemFlags
	wchar_t const* m_name;
	union
	{
		wchar_t const* m_valStr;
		unsigned __int64 m_valNum;
		bool m_valBln;
	};
};

struct SFS_Log
{
	unsigned int m_code;				// SFS_LogAction + SFS_LogStatus
	wchar_t const* m_desc;
	unsigned int m_itemCount;
    SFS_LogItem const* m_itemArray;		// m_itemArray[0-m_itemCount)
};



// SFS context

struct SFS_Context;

typedef bool (__cdecl * SFS_SendResponseType)(SFS_Context* context, SFS_Response const* response);
typedef bool (__cdecl * SFS_SendLogType)(SFS_Context* context, SFS_Log const* log);
typedef bool (__cdecl * SFS_SendLogByHandleType)(SFS_Context* context, unsigned __int64 handle, SFS_Log const* log);
typedef bool (__cdecl * SFS_AddWaitObjectType)(SFS_Context* context, void* waitObject);
typedef bool (__cdecl * SFS_AddTimerType)(SFS_Context* context, unsigned int msFromNow);
typedef bool (__cdecl * SFS_WaitType)(SFS_Context* context);
typedef bool (__cdecl * SFS_SetException)(SFS_Context* context, wchar_t const* what);
typedef bool (__cdecl * SFS_AcquireAsyncContext)(SFS_Context* context, SFS_Context** asyncContext);
typedef bool (__cdecl * SFS_ReleaseAsyncContext)(SFS_Context* asyncContext);


struct SFS_Context
{
	SFS_SendResponseType			SendResponse;
	SFS_SendLogType					SendLog;
	SFS_SendLogByHandleType			SendLogByHandle;
	SFS_AddWaitObjectType			AddWaitObject;
	SFS_AddTimerType				AddTimer;
	SFS_WaitType					Wait;
	SFS_SetException				SetException;
	SFS_AcquireAsyncContext			AcquireAsyncContext;
	SFS_ReleaseAsyncContext			ReleaseAsyncContext;
};



// SFS functions

#ifdef COMPILING_FLOWSFS

typedef __declspec(dllimport) int  (__cdecl * SFS_GetSfsVersionType)();
typedef __declspec(dllimport) bool (__cdecl * SFS_HandleSfsInitType)(unsigned int id, SFS_Init const* init, SFS_Context* context);
typedef __declspec(dllimport) void (__cdecl * SFS_HandleSfsRequestType)(unsigned int id, SFS_Request const* request, SFS_Context* context);
typedef __declspec(dllimport) void (__cdecl * SFS_HandleSfsDestroyType)(unsigned int id, SFS_Context* context);


#else

extern "C" __declspec(dllexport) int  __cdecl GetSfsVersion(); // return SFS_VERSION
extern "C" __declspec(dllexport) bool __cdecl HandleSfsInit(unsigned int id, SFS_Init const* init, SFS_Context* context);
extern "C" __declspec(dllexport) void __cdecl HandleSfsRequest(unsigned int id, SFS_Request const* request, SFS_Context* context);
extern "C" __declspec(dllexport) void __cdecl HandleSfsDestroy(unsigned int id, SFS_Context* context);

#endif
