#include "SfsUtilities.h"

// Windows
#include <Windows.h>



bool SFS_EqualStr(wchar_t const* a, wchar_t const* b, bool caseSensitive)
{
	DWORD lcid = MAKELCID(MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), SORT_DEFAULT);
	return CompareStringW(lcid, caseSensitive ? 0 : NORM_IGNORECASE, a, -1, b, -1) == CSTR_EQUAL;
}


SFS_Init const* SFS_GetNestedTree(SFS_Init const* root, wchar_t const* name)
{
	for (unsigned int i = 0; i < root->m_nestedCount; ++i)
		if (SFS_EqualStr(root->m_nestedArray[i].m_name, name))
			return root->m_nestedArray + i;

	return 0;
}

wchar_t const* SFS_GetNestedValue(SFS_Init const* root, wchar_t const* name)
{
	SFS_Init const* tree = SFS_GetNestedTree(root, name);
	if (tree)
		return tree->m_value;

	return 0;
}


bool SFS_IsYesTrueEnable(wchar_t const* value)
{
	wchar_t const* v[] = { L"Yes", L"True", L"Enable", L"Enabled", L"Available", 0 };
	for (wchar_t const** p = v; *p; ++p)
		if (SFS_EqualStr(value, *p, false))
			return true;

	return false;
}

bool SFS_IsNoFalseDisable(wchar_t const* value)
{
	wchar_t const* v[] = { L"No", L"False", L"Disable", L"Disabled", L"Unavailable", L"Ignore", 0 };
	for (wchar_t const** p = v; *p; ++p)
		if (SFS_EqualStr(value, *p, false))
			return true;

	return false;
}


unsigned int SFS_MatchLists(wchar_t* primaryCsvList, wchar_t const* secondaryCsvList)
{
	DWORD lcid = MAKELCID(MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), SORT_DEFAULT);

	unsigned int matchCount = 0;
	size_t primaryOffset = 0;

	wchar_t const* pBeg = primaryCsvList;
	for (wchar_t const* pEnd = pBeg; *pEnd; pBeg = pEnd + 1)
	{
		for (pEnd = pBeg; *pEnd; ++pEnd)
			if (*pEnd == L',')
				break;

		if (pBeg == pEnd)
			continue;

		wchar_t const* sBeg = secondaryCsvList;
		for (wchar_t const* sEnd = sBeg; *sEnd; sBeg = sEnd + 1)
		{	
			for (sEnd = sBeg; *sEnd; ++sEnd)
				if (*sEnd == L',')
					break;

			if (sBeg == sEnd)
				continue;
				
			if (pEnd - pBeg == sEnd - sBeg)
			{
				size_t size = pEnd - pBeg;
				if (CompareString(lcid, NORM_IGNORECASE, pBeg, (int) size, sBeg, (int) size) == CSTR_EQUAL)
				{
					if (primaryOffset)
						primaryCsvList[primaryOffset++] = L',';

					if (primaryCsvList + primaryOffset != pBeg)
						memmove(primaryCsvList + primaryOffset, pBeg, size * sizeof(wchar_t));
					primaryOffset += size;
					++matchCount;
					break;
				}
			}
		}
	}

	if (primaryOffset)
		primaryCsvList[primaryOffset] = L'\0';

	return matchCount;
}


SFS_LogItem SFS_LogItemGrp(wchar_t const* name, unsigned char flags)
{
	SFS_LogItem item; 
	item.m_type = SFS_LogItemType::Grp;
	item.m_flags = flags;
	item.m_name = name;
	item.m_valNum = 0;
	return item;
}

SFS_LogItem SFS_LogItemNum(wchar_t const* name, unsigned __int64 value, unsigned char flags)
{
	SFS_LogItem item; 
	item.m_type = SFS_LogItemType::Num;
	item.m_flags = flags;
	item.m_name = name;
	item.m_valNum = value;
	return item;
}

SFS_LogItem SFS_LogItemBln(wchar_t const* name, bool value, unsigned char flags)
{
	SFS_LogItem item; 
	item.m_type = SFS_LogItemType::Bln;
	item.m_flags = flags;
	item.m_name = name;
	item.m_valBln = value;
	return item;
}

SFS_LogItem SFS_LogItemStr(wchar_t const* name, wchar_t const* value, unsigned char flags)
{
	SFS_LogItem item; 
	item.m_type = SFS_LogItemType::Str;
	item.m_flags = flags;
	item.m_name = name;
	item.m_valStr = value;
	return item;
}


wchar_t const* SFS_ToString_ControlByte(unsigned char controlByte)
{
	switch (controlByte)
    {
	case SFS_ControlByte::NoCheck:		return L"NoCheck";
	case SFS_ControlByte::StatIf:		return L"StatIf";
	case SFS_ControlByte::StatAlways:	return L"StatAlways";
	default:							return 0;
	}
}

wchar_t const* SFS_ToString_RequestType(unsigned int requestType)
{
    switch (requestType)
    {
    case SFS_RequestType::Open:				return L"Open";
    case SFS_RequestType::Close:			return L"Close";
    case SFS_RequestType::Read:				return L"Read";
    case SFS_RequestType::Write:			return L"Write";
    case SFS_RequestType::FStat:			return L"FStat";
    case SFS_RequestType::SetStat:			return L"SetStat";
    case SFS_RequestType::FSetStat:			return L"FSetStat";
    case SFS_RequestType::OpenDir:			return L"OpenDir";
    case SFS_RequestType::ReadDir:			return L"ReadDir";
    case SFS_RequestType::Remove:			return L"Remove";
    case SFS_RequestType::MkDir:			return L"MkDir";
    case SFS_RequestType::RmDir:			return L"RmDir";
    case SFS_RequestType::RealPath:			return L"RealPath";
    case SFS_RequestType::Stat:				return L"Stat";
    case SFS_RequestType::Rename:			return L"Rename";
    case SFS_RequestType::ReadLink:			return L"ReadLink";
    case SFS_RequestType::Link:				return L"Link";
    case SFS_RequestType::Block:			return L"Block";
    case SFS_RequestType::Unblock:			return L"Unblock";
    case SFS_RequestType::TextSeek:			return L"TextSeek";
    case SFS_RequestType::CheckFileHandle:	return L"CheckFileHandle";
    case SFS_RequestType::CheckFileName:	return L"CheckFileName";
    case SFS_RequestType::SpaceAvail:		return L"SpaceAvail";
	case SFS_RequestType::FileCopy:			return L"FileCopy";
	case SFS_RequestType::DataCopy:			return L"DataCopy";
	case SFS_RequestType::NativePath:		return L"NativePath";
    default:								return 0;
    }
}

wchar_t const* SFS_ToString_ResponseType(unsigned int responseType)
{
    switch (responseType)
    {
	case SFS_ResponseType::Status:			return L"Status";
	case SFS_ResponseType::Handle:			return L"Handle";
    case SFS_ResponseType::Data:			return L"Data";
    case SFS_ResponseType::Names:			return L"Names";
    case SFS_ResponseType::Attrs:			return L"Attrs";
    case SFS_ResponseType::Name:			return L"Name";
    case SFS_ResponseType::CheckFileReply:	return L"CheckFileReply";
	default:								return 0;
	}
}

wchar_t const* SFS_ToString_StatusCode(unsigned int statusCode)
{
	switch (statusCode)
    {
    case SFS_StatusCode::Ok:						return L"Ok";
    case SFS_StatusCode::Eof:						return L"Eof";
    case SFS_StatusCode::NoSuchFile:				return L"NoSuchFile";
	case SFS_StatusCode::PermissionDenied:			return L"PermissionDenied";
    case SFS_StatusCode::Failure:					return L"Failure";
    case SFS_StatusCode::BadMessage:				return L"BadMessage";
    case SFS_StatusCode::NoConnection:				return L"NoConnection";
	case SFS_StatusCode::ConnectionLost:			return L"ConnectionLost";
    case SFS_StatusCode::OpUnsupported:				return L"OpUnsupported";
    case SFS_StatusCode::InvalidHandle:				return L"InvalidHandle";
    case SFS_StatusCode::NoSuchPath:				return L"NoSuchPath";
    case SFS_StatusCode::FileAlreadyExists:			return L"FileAlreadyExists";
    case SFS_StatusCode::WriteProtect:				return L"WriteProtect";
    case SFS_StatusCode::NoMedia:					return L"NoMedia";
	case SFS_StatusCode::NoSpaceOnFilesystem:		return L"NoSpaceOnFilesystem";
    case SFS_StatusCode::QuotaExceeded:				return L"QuotaExceeded";
    case SFS_StatusCode::UnknownPrincipal:			return L"UnknownPrincipal";
    case SFS_StatusCode::LockConflict:				return L"LockConflict";
    case SFS_StatusCode::DirNotEmpty:				return L"DirNotEmpty";
    case SFS_StatusCode::NotADirectory:				return L"NotADirectory";
    case SFS_StatusCode::InvalidFilename:			return L"InvalidFileName";
    case SFS_StatusCode::LinkLoop:					return L"LinkLoop";
    case SFS_StatusCode::CannotDelete:				return L"CannotDelete";
    case SFS_StatusCode::InvalidParameter:			return L"InvalidParameter";
    case SFS_StatusCode::FileIsADirectory:			return L"FileIsADirectory";
    case SFS_StatusCode::ByteRangeLockConflict:		return L"ByteRangeLockConflict";
    case SFS_StatusCode::ByteRangeLockRefused:		return L"ByteRangeLockRefused";
	case SFS_StatusCode::DeletePending:				return L"DeletePending";
	case SFS_StatusCode::FileCorrupt:				return L"FileCorrupt";
    case SFS_StatusCode::OwnerInvalid:				return L"OwnerInvalid";
    case SFS_StatusCode::GroupInvalid:				return L"GroupInvalid";
    case SFS_StatusCode::NoMatchingByteRangeLock:	return L"NoMatchingByteRangeLock";
	default:										return 0;
    }
}

wchar_t const* SFS_Describe_StatusCode(unsigned int statusCode)
{
	switch (statusCode)
    {
    case SFS_StatusCode::Ok:						return L"The operation completed successfully.";
    case SFS_StatusCode::Eof:						return L"Reached the end of the file.";
    case SFS_StatusCode::NoSuchFile:				return L"The file does not exist.";
	case SFS_StatusCode::PermissionDenied:			return L"Permission denied.";
    case SFS_StatusCode::Failure:					return L"The requested operation failed.";
    case SFS_StatusCode::BadMessage:				return L"A bad message received.";
    case SFS_StatusCode::NoConnection:				return L"There is no connection to the server.";
	case SFS_StatusCode::ConnectionLost:			return L"The connection to the server was lost.";
    case SFS_StatusCode::OpUnsupported:				return L"The requested operation is not supported.";
    case SFS_StatusCode::InvalidHandle:				return L"The handle is invalid.";
    case SFS_StatusCode::NoSuchPath:				return L"The file path does not exist or is invalid.";
    case SFS_StatusCode::FileAlreadyExists:			return L"The file already exists.";
    case SFS_StatusCode::WriteProtect:				return L"The media is write protected.";
    case SFS_StatusCode::NoMedia:					return L"No media in drive.";
	case SFS_StatusCode::NoSpaceOnFilesystem:		return L"Insufficient free space on the filesystem.";
    case SFS_StatusCode::QuotaExceeded:				return L"Not enough quota is available.";
    case SFS_StatusCode::UnknownPrincipal:			return L"A principal referenced by the request is unknown.";
    case SFS_StatusCode::LockConflict:				return L"The file could not be opened because it is locked by another process.";
    case SFS_StatusCode::DirNotEmpty:				return L"The directory is not empty.";
    case SFS_StatusCode::NotADirectory:				return L"The specified file is not a directory.";
    case SFS_StatusCode::InvalidFilename:			return L"The filename is not valid.";
    case SFS_StatusCode::LinkLoop:					return L"Too many symbolic links encountered.";
    case SFS_StatusCode::CannotDelete:				return L"The file cannot be deleted.";
    case SFS_StatusCode::InvalidParameter:			return L"The parameter is incorrect.";
    case SFS_StatusCode::FileIsADirectory:			return L"The specified file is a directory.";
    case SFS_StatusCode::ByteRangeLockConflict:		return L"Another process's mandatory byte-range lock overlaps with this request.";
    case SFS_StatusCode::ByteRangeLockRefused:		return L"A request for a byte range lock was refused.";
	case SFS_StatusCode::DeletePending:				return L"An operation was attempted on a file for which a delete operation is pending.";
	case SFS_StatusCode::FileCorrupt:				return L"The file is corrupt.";
    case SFS_StatusCode::OwnerInvalid:				return L"The principal specified can not be assigned as an owner of a file.";
    case SFS_StatusCode::GroupInvalid:				return L"The principal specified can not be assigned as the primary group of a file.";
    case SFS_StatusCode::NoMatchingByteRangeLock:	return L"The requested operation could not be completed because the specifed byte range lock has not been granted.";
	default:										return 0;
    }
}



namespace SFS
{


template <class T> T Max(T first, T second)
{
	return first > second ? first : second;
}


class AutoLocalFree
{
public:
	AutoLocalFree(void* buffer) : m_buffer(buffer) {}
	~AutoLocalFree() { LocalFree(m_buffer); }
private:
	void* m_buffer;
};


TransferFile::TransferFile()
	: m_createdNewFile(false)
	, m_resizedFile(false)
{
	GetSystemTimeAsFileTime((FILETIME*) &m_startTime);
}

void TransferFile::RegisterReadEof()
{
	if (m_read.m_count < (unsigned int) -1)
		++m_read.m_count;
}

unsigned int TransferFile::RegisterCount() const
{
	unsigned __int64 count = 0;
	count += m_read.m_count;
	count += m_write.m_count;
	if (m_createdNewFile)
		++count;
	if (m_resizedFile)
		++count;

	if (count < (unsigned int) -1)
		return (unsigned int) count;
	else
		return (unsigned int) -1;	
}

unsigned __int64 TransferFile::TimeMs() const
{
	unsigned __int64 endTime;
	GetSystemTimeAsFileTime((FILETIME*) &endTime);
	return (endTime - m_startTime) / 10000;
}

TransferFile::Data::Data() 
{
	ZeroMemory(this, sizeof(*this));
}

void TransferFile::Data::Register(unsigned __int64 offset, unsigned __int64 length)
{
	if (m_count < (unsigned int) -1)
		++m_count;

	if (length)
	{
		m_bytes += length;

		unsigned int replaceIdx = 0;
		unsigned int matchCount = 0;
		
		for (unsigned idx = 0; idx < MaxRangeTrack; ++idx)
		{
			Range& r = m_ranges[idx];

			if (r.m_length)
			{
				if (offset == -1 && r.m_offset == -1) // if appending
				{
					++matchCount;
					r.m_length += length;
					break; // there cannot be any overlapping when appending
				}
				else if (offset < r.m_offset && (offset + length) >= r.m_offset)
				{
					++matchCount;
					r.m_length = Max(length, (r.m_offset - offset) + r.m_length);
					r.m_offset = offset;
				}
				else if (offset >= r.m_offset && offset <= (r.m_offset + r.m_length))
				{
					++matchCount;
					r.m_length = (r.m_length, (offset - r.m_offset) + length);
				}
			}

			if (idx)
				if (r.m_length < m_ranges[replaceIdx].m_length)
					replaceIdx = idx;
		}

		if (matchCount == 0)
		{
			Range& r = m_ranges[replaceIdx];

			r.m_offset = offset;
			r.m_length = length;
		}
		else if (matchCount > 1) // ranges need to be merged
		{
			for (unsigned idxA = 0; idxA < MaxRangeTrack - 1; ++idxA)
			{
				Range& ra = m_ranges[idxA];
				if (!ra.m_length)
					continue;

				for (unsigned idxB = idxA + 1; idxB < MaxRangeTrack; ++idxB)
				{
					Range& rb = m_ranges[idxB];
					if (!rb.m_length)
						continue;

					if (ra.m_offset < rb.m_offset && (ra.m_offset + ra.m_length) >= rb.m_offset)
					{
						ra.m_length = Max(ra.m_length, (rb.m_offset - ra.m_offset) + rb.m_length);

						rb.m_offset = 0;
						rb.m_length = 0;
					}
					else if (ra.m_offset >= rb.m_offset && ra.m_offset <= (rb.m_offset + rb.m_length))
					{
						ra.m_length = Max(rb.m_length, (ra.m_offset - rb.m_offset) + ra.m_length);				
						ra.m_offset = rb.m_offset;

						rb.m_offset = 0;
						rb.m_length = 0;
					}
				}
			}
		}
	}
}

TransferFile::Range const& TransferFile::Data::LargestRange() const
{
	unsigned returnIdx = 0;

	for (unsigned idx = 1; idx < MaxRangeTrack; ++idx)
		if (m_ranges[idx].m_length > m_ranges[returnIdx].m_length)
			returnIdx = idx;
	
	return m_ranges[returnIdx];
}


ListDir::ListDir() 
	: m_count(0)
	, m_entries(0)
{
	GetSystemTimeAsFileTime((FILETIME*) &m_startTime);
}

void ListDir::RegisterList(unsigned int entries)
{ 
	if (m_count < (unsigned int) -1)
		++m_count; 

	unsigned __int64 i = 0;
	i += m_entries;
	i += entries;
	if (i < (unsigned int) -1)
		m_entries = (unsigned int) i;
	else
		m_entries = (unsigned int) -1;
}

void ListDir::RegisterListEof()
{ 
	if (m_count < (unsigned int) -1)
		++m_count; 
}

unsigned __int64 ListDir::TimeMs() const
{
	unsigned __int64 endTime;
	GetSystemTimeAsFileTime((FILETIME*) &endTime);
	return (endTime - m_startTime) / 10000;
}


Log::Log(DefaultCode code, unsigned int reserve)
{
	m_log.m_code = code;
	m_log.m_desc = 0;
	m_items.Reserve(reserve);
}

Log::Log(unsigned int code, wchar_t const* desc, bool copyDesc, unsigned int reserve)
{
	m_log.m_code = code;
	m_log.m_desc = copyDesc ? m_strings.Copy(desc) : desc;
	m_items.Reserve(reserve);
}

Log::Log(wchar_t const* path, bool copyPath, ListDir const& dir)
{
	m_log.m_code = SFS_LogCode::ListDirDone; 
	m_log.m_desc = 0;
	m_items.Reserve(4);
	
	AddGrp_Param();
	AddStr(L"path", path, copyPath ? CopyValue : 0);
	AddNum(L"timeMs", dir.TimeMs());
	AddNum(L"entiresCount", dir.EntiresCount());
}

Log::Log(wchar_t const* path, bool copyPath, TransferFile const& file)
{
	m_log.m_code = SFS_LogCode::TransferFileDone;
	m_log.m_desc = 0;
	m_items.Reserve(9);
	
	AddGrp_Param();
	AddStr(L"path", path, copyPath ? CopyValue : 0);
	AddNum(L"timeMs", file.TimeMs());
	AddNum(L"bytesRead", file.BytesRead());
	AddNum(L"bytesWritten", file.BytesWritten());
	TransferFile::Range const& readRange = file.LargestReadRange();
	AddNum(L"readRangeOffset", readRange.m_offset);
	AddNum(L"readRangeLength", readRange.m_length);
	TransferFile::Range const& writeRange = file.LargestWriteRange();
	if (writeRange.m_offset == -1)
		AddStr(L"writeRangeOffset", L"Eof");
	else
		AddNum(L"writeRangeOffset", writeRange.m_offset);
	AddNum(L"writeRangeLength", writeRange.m_length);
	AddBln(L"createdNewFile", file.CreatedNewFile());
	AddBln(L"resizedFile", file.ResizedFile());
}


Log& Log::AddGrp(wchar_t const* name, unsigned char flags)
{
	m_items.PushBack(SFS_LogItemGrp(flags & CopyName ? m_strings.Copy(name) : name, flags));	
	return *this;
}

Log& Log::AddNum(wchar_t const* name, unsigned __int64 value, unsigned char flags)
{
	m_items.PushBack(SFS_LogItemNum(flags & CopyName ? m_strings.Copy(name) : name, value, flags));	
	return *this;
}

Log& Log::AddBln(wchar_t const* name, bool value, unsigned char flags)
{
	m_items.PushBack(SFS_LogItemBln(flags & CopyName ? m_strings.Copy(name) : name, value, flags));	
	return *this;
}

Log& Log::AddStr(wchar_t const* name, wchar_t const* value, unsigned char flags)
{
	m_items.PushBack(SFS_LogItemStr(
		flags & CopyName ? m_strings.Copy(name) : name,
		flags & CopyValue ? m_strings.Copy(value) : value,
		flags));	
	return *this;
}

Log& Log::AddGrp_Error_WinApi(wchar_t const* operation, bool copyOperation, unsigned int code)
{
	AddReserve(5);
	AddGrp_Error();
	AddStr(L"type", L"WinAPI");
	AddStr(L"operation", operation, copyOperation ? CopyValue : 0);
	AddNum(L"code", code, (code <= 99999 ? 0 : Hex));

	wchar_t* message;
	DWORD flags = FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS;
	if (!FormatMessageW(flags, 0, code, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT), (wchar_t*) &message, 0, 0))
		return AddStr(L"message", L"[Error retrieving WinAPI message]");
	else
	{
		AutoLocalFree autoLocalFree(message);

		// Remove CR and LF at the end of the message
		int size = lstrlenW(message);
		while (size > 0)
		{
			if (message[size-1] == L'\n')
				message[--size] = L'\0';
			else if (message[size-1] == L'\r')
				message[--size] = L'\0';
			else
				break;
		}

		return AddStr(L"message", message, CopyValue);
	}
}

Log& Log::AddGrp_Error_Exception(wchar_t const* what, bool copyWhat)
{
	AddReserve(3);
	AddGrp_Error();
	AddStr(L"type", L"Exception");

	byte whatOptions = copyWhat ? CopyValue : 0;
	for (wchar_t const* c = what; *c; ++c)
	{
		if (*c == L'\r' || *c == L'\n')
		{
			whatOptions |= MultiLine;
			break;
		}
	}

	return AddStr(L"message", what, whatOptions);
}

Log& Log::AddGrp_Error_Exception(char const* what)
{
	wchar_t const* wide = L"[Error converting exception message]";

	int whatLen = lstrlenA(what);
	if (!whatLen)
		wide = L"";
	else
	{
		DWORD flags = MB_PRECOMPOSED | MB_ERR_INVALID_CHARS;
		int c = MultiByteToWideChar(CP_ACP, flags, what, whatLen, 0, 0);
		if (!c && GetLastError() == ERROR_INVALID_FLAGS)
			c = MultiByteToWideChar(CP_ACP, flags = 0, what, whatLen, 0, 0);

		if (c)
		{
			wchar_t* buffer = m_strings.Create(c + 1);
			buffer[c] = L'\0';
			MultiByteToWideChar(CP_ACP, flags, what, whatLen, buffer, c);
		}
	}

	return AddGrp_Error_Exception(wide, false);
}

Log& Log::AddGrp_Error_Exception()
{
	AddReserve(2);
	return AddGrp_Error().AddStr(L"type", L"Exception");
}

Log& Log::AddGrp_Error_StatusCode(unsigned int statusCode)
{
	wchar_t const* what = SFS_Describe_StatusCode(statusCode);
	if (!what) 
		what = SFS_Describe_StatusCode(SFS_StatusCode::Failure);
	return AddGrp_Error_Exception(what, false);
}

Log& Log::AddStatusCode(unsigned int statusCode)
{
	wchar_t const* str = SFS_ToString_StatusCode(statusCode);
	if (str)
		return AddStr(L"statusCode", str);
	else
		return AddNum(L"statusCode", statusCode);
}

Log& Log::AddControlByte(unsigned char controlByte) 
{
	wchar_t const* str = SFS_ToString_ControlByte(controlByte);
	if (str)
		return AddStr(L"controlByte", str);
	else
		return AddNum(L"controlByte", controlByte);
}

Log& Log::AddRequestType(unsigned int requestType) 
{
	wchar_t const* str = SFS_ToString_RequestType(requestType);
	if (str)
		return AddStr(L"requestType", str);
	else
		return AddNum(L"requestType", requestType);
}

Log& Log::AddGrp_Help(wchar_t const* message, unsigned char messageOptions)
{
	AddReserve(2);
	return AddGrp_Help().AddStr(L"message", message, messageOptions);
}

Log::operator SFS_Log const*()
{
	m_log.m_itemCount = (unsigned int) m_items.Size();
	m_log.m_itemArray = m_items.Data();
    return &m_log;
}

void Log::Items::Reserve(size_t capacity)
{
	if (m_capacity < capacity)
	{
		SFS_LogItem* data = new SFS_LogItem[capacity];
		CopyMemory(data, m_data, m_size * sizeof(SFS_LogItem));
		delete [] m_data;
		m_data = data;
		m_capacity = capacity;
	}
}

void Log::Items::PushBack(SFS_LogItem const& item)
{
	if (m_capacity == m_size)
		Reserve(m_capacity ? m_capacity * 2 : 4);
	m_data[m_size++] = item;
}


Log::Strings::~Strings()
{
	while (m_list)
	{
		Node* next = m_list->m_next;
		delete [] m_list;
		m_list = next;
	}
}

wchar_t* Log::Strings::Create(size_t chars)
{
	size_t sizeOfStr = (chars + 1) * sizeof(wchar_t);
	Node* node = (Node*) new unsigned char[sizeof(Node) + sizeOfStr];
	node->m_next = m_list;
	m_list = node;
	
	wchar_t* str = (wchar_t*) (node + 1);
	str[chars] = L'\0';
	return str;
}

wchar_t* Log::Strings::Copy(wchar_t const* str)
{
	size_t chars = (lstrlenW(str) + 1);	
	wchar_t* strCopy = Create(chars);
	CopyMemory(strCopy, str, chars * sizeof(wchar_t));
	return strCopy;
}


}