#pragma once

#include "SfsProvider.h"


bool SFS_EqualStr(wchar_t const* a, wchar_t const* b, bool caseSensitive = true);

SFS_Init const* SFS_GetNestedTree(SFS_Init const* root, wchar_t const* name);
wchar_t const* SFS_GetNestedValue(SFS_Init const* root, wchar_t const* name);

bool SFS_IsYesTrueEnable(wchar_t const* value);
bool SFS_IsNoFalseDisable(wchar_t const* value);

// Returns the number of matched items while the matching list is stored in the first parameter.
unsigned int SFS_MatchLists(wchar_t* primaryCsvList, wchar_t const* secondaryCsvList);

SFS_LogItem SFS_LogItemBrk();
SFS_LogItem SFS_LogItemGrp(wchar_t const* name, unsigned char flags = 0);
SFS_LogItem SFS_LogItemNum(wchar_t const* name, unsigned __int64 value, unsigned char flags = 0);
SFS_LogItem SFS_LogItemBln(wchar_t const* name, bool value, unsigned char flags = 0);
SFS_LogItem SFS_LogItemStr(wchar_t const* name, wchar_t const* value, unsigned char flags = 0);

// NULL is returned if the passed value is unrecognized.
wchar_t const* SFS_ToString_ControlByte(unsigned char controlByte);
wchar_t const* SFS_ToString_RequestType(unsigned int requestType);
wchar_t const* SFS_ToString_ResponseType(unsigned int responseType);
wchar_t const* SFS_ToString_StatusCode(unsigned int statusCode);
wchar_t const* SFS_Describe_StatusCode(unsigned int statusCode);


namespace SFS
{


class TransferFile
{
public:
	TransferFile();
	
	void RegisterCreateNewFile()											{ m_createdNewFile = true; }
	void RegisterResizeFile()												{ m_resizedFile = true; }
	void RegisterWrite(unsigned __int64 offset, unsigned __int64 length)	{ m_write.Register(offset, length); }
	void RegisterRead(unsigned __int64 offset, unsigned __int64 length)		{ m_read.Register(offset, length); }
	void RegisterReadEof();
	
	operator SFS_Log const*();

	struct Range
	{
		unsigned __int64 m_offset;
		unsigned __int64 m_length;
	};

	unsigned int RegisterCount() const;
	bool CreatedNewFile() const				{ return m_createdNewFile; }
	bool ResizedFile() const				{ return m_resizedFile; }
	unsigned __int64 BytesRead() const		{ return m_read.m_bytes; }
	unsigned __int64 BytesWritten() const	{ return m_write.m_bytes; }
	Range const& LargestReadRange() const	{ return m_read.LargestRange(); }
	Range const& LargestWriteRange() const	{ return m_write.LargestRange(); }
	unsigned __int64 TimeMs() const;

private:
	class Data
	{
	public:
		Data();
		
		void Register(unsigned __int64 offset, unsigned __int64 length);
		Range const& LargestRange() const;

		unsigned int m_count;
		unsigned __int64 m_bytes;
		enum { MaxRangeTrack = 8 };		// must be at least 1
		Range m_ranges[MaxRangeTrack];	// using 16 bytes per track
	};

	Data m_read;
	Data m_write;
	bool m_createdNewFile;
	bool m_resizedFile;
	unsigned __int64 m_startTime;
};


class ListDir
{
public:
	ListDir();

	void RegisterList(unsigned int entries);
	void RegisterListEof();

	unsigned int RegisterCount() const		{ return m_count; } 
	unsigned int EntiresCount() const		{ return m_entries; }
	unsigned __int64 TimeMs() const;

private:
	unsigned int m_count;
	unsigned int m_entries;
	unsigned __int64 m_startTime;

};


class Log
{
public:
	enum DefaultCode { Success = 0, Failure = 200, Progress = 600 };
	Log(DefaultCode code, unsigned int reserve = 0);
	Log(unsigned int code, wchar_t const* desc, bool copyDesc, unsigned int reserve = 0);
	Log(wchar_t const* path, bool copyPath, ListDir const& dir);
	Log(wchar_t const* path, bool copyPath, TransferFile const& file);
	virtual ~Log() {}

	Log& AddReserve(size_t reserve) { return m_items.Reserve(m_items.Size() + reserve), *this; }

	enum Flags { CopyName = 1, CopyValue = 2, MultiLine = 4, NewLevel = 8, Hex = 16 };

	Log& AddGrp(wchar_t const* name, unsigned char flags = 0);							// valid flags: CopyName, NewLevel
	Log& AddNum(wchar_t const* name, unsigned __int64 value, unsigned char flags = 0);	// valid flags: CopyName, Hex
	Log& AddBln(wchar_t const* name, bool value, unsigned char flags = 0);				// valid flags: CopyName
	Log& AddStr(wchar_t const* name, wchar_t const* value, unsigned char flags = 0);	// valid flags: CopyName, CopyValue, MultiLine
	
	Log& AddGrp_Param() { return AddGrp(L"parameters"); }
	Log& AddGrp_Error()	{ return AddGrp(L"error"); }
	Log& AddGrp_Help()	{ return AddGrp(L"help"); }

	Log& AddGrp_Error_WinApi(wchar_t const* operation, bool copyOperation, unsigned int code);	// 5 items
	Log& AddGrp_Error_Exception(wchar_t const* what, bool copyWhat);							// 3 items
	Log& AddGrp_Error_Exception(char const* what);												// 3 items
	Log& AddGrp_Error_Exception(); // unrecognized exception									// 2 items
	Log& AddGrp_Error_StatusCode(unsigned int statusCode); // logged as an exception			// 3 items
	
	Log& AddStatusCode(unsigned int statusCode);											// 1 item
	Log& AddControlByte(unsigned char controlByte);											// 1 item
	Log& AddRequestType(unsigned int requestType);											// 1 item

	Log& AddGrp_Help(wchar_t const* message, unsigned char messageOptions = 0); // valid flags:  CopyValue, MultiLine - 2 items

	operator SFS_Log const*();
	
private:
	class Items
	{
	public:
		Items() : m_capacity(0), m_size(0), m_data(0) {}
		~Items() { if (m_data) delete [] m_data; }

		void Reserve(size_t capacity);
		void PushBack(SFS_LogItem const& item);
		SFS_LogItem const* Data() const { return m_data; }
		size_t Size() const { return m_size; }

	private:
		size_t m_capacity;
		size_t m_size;
		SFS_LogItem* m_data;
	};

	class Strings
	{
	public:
		Strings() : m_list(0) {}
		~Strings();

		wchar_t* Create(size_t chars);
		wchar_t* Copy(wchar_t const* str);

	private:
		struct Node	{ Node* m_next; } *m_list;
	};

	Items m_items;
	Strings m_strings;
	SFS_Log m_log;
};


}