// SfsWin
#include "SfsWin.h"

// Windows
#include <Sddl.h>

// CRT
#include <string.h>

// STL
#include <limits>
#include <sstream>


using namespace std;
using namespace SFS;


unsigned int const SfsWin::s_maxReadSize = 5*1024*1024; // 5MB


SfsWin::~SfsWin() 
{ 
	if (m_overlapped.hEvent) 
		CloseHandle(m_overlapped.hEvent); 

	for (FsObjMap::iterator it = m_fsObjs.begin(); it != m_fsObjs.end(); ++it)
		delete it->second;
}

bool SfsWin::HandleInit(SFS_Init const* init, SFS_Context* context)
{
	m_sfsContext = context;

	try
	{
		try
		{
			wchar_t const* value = SFS_GetNestedValue(init, L"MountPath");
			if (value)
				m_mountPath = value;

			// This sample does not support RootPath or any other custom
			// provider settings. Consider the following sample code:
			//
			// value = SFS_GetNestedValue(init, L"RootPath");
			// if (value)
			//     m_rootPath = value;
			//
			// m_showHidden = true;
			// value = GetNestedValue(init->m_root, L"ShowHidden");
			// if (value && SFS_IsNoFalseDisable(value))
			//     m_showHidden = false;
			
			m_overlapped.hEvent = CreateEvent(0, true, false, 0);
			if (!m_overlapped.hEvent)
			{
				DWORD winError = GetLastError();

				SfsSendLog(Log(Log::Failure, 5)
					.AddGrp_Error_WinApi(L"CreateEvent()", false, winError));

				return false;
			}
		}
		catch (WException const& e)
		{
			SfsSetException(e.What());
		}
		catch (std::exception const& e)
		{
			wstring wide;
			if (!NarrowToWide(e.what(), CP_ACP, wide))
				SfsSetException(L"SfsWin::HandleInit() handler failed: [error converting std::exception message to Unicode]");
			else
			{
				wstring s;
				s.reserve(50 + wide.size());
				s.append(L"SfsWin::HandleInit() handler failed: ");
				s.append(wide);
				SfsSetException(s.c_str());
			}
		}
	}
	catch (std::bad_alloc const&)
	{
		SfsSetException(L"SfsWin::HandleInit() handler failed: bad allocation");
	}
	catch (SfsContextFailure const&)
	{
	}
	
	return true;
}

void SfsWin::HandleRequest(SFS_Request const* request, SFS_Context* context)
{
	m_sfsContext = context;

	try
	{
		try
		{
			switch (request->m_type)
			{
			default:
				SfsSendLog(Log(Log::Failure, 3 + 3)
					.AddGrp_Param().AddRequestType(request->m_type)
								   .AddStatusCode(SFS_StatusCode::OpUnsupported)
					.AddGrp_Error_Exception(L"Unsupported request received.", false));	

				SfsSendResponse(SfsStatusResponse(SFS_StatusCode::OpUnsupported));
				break;

			case SFS_RequestType::Open:		HandleOpen(request);		break;
			case SFS_RequestType::Close:	HandleClose(request);		break;
			case SFS_RequestType::Read:		HandleRead(request);		break;
			case SFS_RequestType::Write:	HandleWrite(request);		break;
			case SFS_RequestType::Stat:		
			case SFS_RequestType::FStat:	HandleStat(request);		break;
			case SFS_RequestType::SetStat:
			case SFS_RequestType::FSetStat:	HandleSetStat(request);		break;
			case SFS_RequestType::OpenDir:	HandleOpenDir(request);		break;
			case SFS_RequestType::ReadDir:	HandleReadDir(request);		break;
			case SFS_RequestType::Remove:	HandleRemove(request);		break;
			case SFS_RequestType::MkDir:	HandleMkDir(request);		break;
			case SFS_RequestType::RmDir:	HandleRmDir(request);		break;
			case SFS_RequestType::RealPath:	HandleRealPath(request);	break;
			case SFS_RequestType::Rename:	HandleRename(request);		break;
			case SFS_RequestType::ReadLink:	HandleReadLink(request);	break;
			case SFS_RequestType::Link:		HandleLink(request);		break;
			case SFS_RequestType::Block:	HandleBlock(request);		break;
			case SFS_RequestType::Unblock:	HandleUnblock(request);		break;
			case SFS_RequestType::TextSeek:	HandleTextSeek(request);	break;
			case SFS_RequestType::CheckFileName:
			case SFS_RequestType::CheckFileHandle:	HandleCheckFile(request);	break;
			case SFS_RequestType::SpaceAvail:		HandleSpaceAvail(request);	break;
			case SFS_RequestType::FileCopy:			HandleFileCopy(request);	break;
			case SFS_RequestType::DataCopy:			HandleDataCopy(request);	break;
			case SFS_RequestType::NativePath:		HandleNativePath(request);	break;
			}
		}
		catch (SfsStatusResponse const& sfsStatusResponse)
		{
			SfsSendResponse(sfsStatusResponse);
		}
		catch (WException const& e)
		{
			SfsSetException(e.What());
		}
		catch (std::exception const& e)
		{
			wstring wide;
			if (!NarrowToWide(e.what(), CP_ACP, wide))
				SfsSetException(L"SfsWin::HandleRequest() handler failed: [error converting std::exception message to Unicode]");
			else
			{
				wstring s;
				s.reserve(50 + wide.size());
				s.append(L"SfsWin::HandleRequest() handler failed: ");
				s.append(wide);
				SfsSetException(s.c_str());
			}
		}
	}
	catch (std::bad_alloc const&)
	{
		SfsSetException(L"SfsWin::HandleRequest() handler failed: bad allocation");
	}
	catch (SfsContextFailure const&)
	{
	}
}

void SfsWin::HandleDestroy(SFS_Context* context)
{
	m_sfsContext = context;

	try
	{
		try
		{
			for (FsObjMap::iterator it = m_fsObjs.begin(); it != m_fsObjs.end(); ++it)
			{
				FsObj* obj = it->second;

				if (obj->Root())
				{
					FsRoot* root = obj->Root();
					if (root->m_listDir.RegisterCount())
						SfsSendLogByHandle(obj->m_sfsHandle, Log(obj->m_winPath.c_str(), false, root->m_listDir));
				}
				else if (obj->Dir())
				{
					FsDir* dir = obj->Dir();
					dir->CloseFindHandle();

					if (dir->m_listDir.RegisterCount())
						SfsSendLogByHandle(obj->m_sfsHandle, Log(obj->m_winPath.c_str(), false, dir->m_listDir));
				}
				else // if (obj->File())
				{
					FsFile* file = obj->File();
					file->CloseFileHandle();

					if (file->m_transferFile.RegisterCount())
						SfsSendLogByHandle(obj->m_sfsHandle, Log(obj->m_winPath.c_str(), false, file->m_transferFile));
				}
									
				SfsSendLogByHandle(obj->m_sfsHandle, Log(SFS_LogCode::CloseSuccess, 0, false, 3)
					.AddGrp_Param().AddStr(L"path", obj->m_winPath.c_str())
								   .AddStr(L"type", obj->File() ? L"File" : L"Directory"));
			}
		}
		catch (WException const& e)
		{
			SfsSetException(e.What());
		}
		catch (std::exception const& e)
		{
			wstring wide;
			if (!NarrowToWide(e.what(), CP_ACP, wide))
				SfsSetException(L"SfsWin::HandleDestroy() handler failed: [error converting std::exception message to Unicode]");
			else
			{
				wstring s;
				s.reserve(50 + wide.size());
				s.append(L"SfsWin::HandleDestroy() handler failed: ");
				s.append(wide);
				SfsSetException(s.c_str());
			}
		}
	}
	catch (std::bad_alloc const&)
	{
		SfsSetException(L"SfsWin::HandleDestroy() handler failed: bad allocation");
	}
	catch (SfsContextFailure const&)
	{
	}
}


void SfsWin::HandleOpen(SFS_Request const* request)
{
	SFS_Open const* open = (SFS_Open const*) request->m_content;

	auto_ptr<FsFile> autoFile(new FsFile);
	FsFile* file = autoFile.get();
	
	file->m_sfsPath = open->m_fileName;
	ExtractWinPath(request, file->m_winPath);

	if (open->m_flags & SFS_OpenFlag::AppendAtomic)
	{
		if (!(open->m_flags & SFS_OpenFlag::Append) && !(open->m_flags & SFS_OpenFlag::BlockWrite))
		{
			wchar_t const* reason = L"The AppendAtomic flag is not supported.";
			
			SfsSendLog(Log(Log::Failure, 4 + 3)
				.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
							   .AddNum(L"flags", open->m_flags)
							   .AddStatusCode(SFS_StatusCode::Failure)
				.AddGrp_Error_Exception(reason, false));

			throw SfsStatusResponse(SFS_StatusCode::Failure, reason);
		}
	}

	SFS_Attrs const* sfsAttrs = ExtractSfsAttrs(request);

	WinTimes winTimes;
	ExtractWinTimes(request, winTimes);

	// At this point, the built-in FlowSfsWin would also extract security
	// attributes (owner, group, POSIX permissions, ACL) which it would 
	// then pass on to the CreateFile() call.

	DWORD winAttrs = GetFileAttributes(file->m_winPath.c_str());
	if (winAttrs != INVALID_FILE_ATTRIBUTES)
	{
		if (winAttrs & FILE_ATTRIBUTE_DIRECTORY)
		{
			unsigned int statusCode = SFS_StatusCode::FileIsADirectory;
			SfsSendLog(Log(Log::Failure, 3 + 3)
				.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
							   .AddStatusCode(statusCode)
				.AddGrp_Error_StatusCode(statusCode));

			throw SfsStatusResponse(statusCode);
		}

		if (open->m_flags && SFS_OpenFlag::NoFollow)
		{
			// If file->m_winPath is a symbolic link, then at this point,
			// the build-in FlowSfsWin would fail this request by sending
			// a SFS_StatusCode::LinkLoop status response.
		}
	}

	DWORD creatDisp = (open->m_flags & SFS_OpenFlag::AccessDispositionMask) + 1;
			
	file->m_append = !!(open->m_flags & (SFS_OpenFlag::Append | SFS_OpenFlag::AppendAtomic));
	file->m_textMode = !!(open->m_flags & SFS_OpenFlag::TextMode);

	DWORD shareMode = 0;
	if (!(open->m_flags & SFS_OpenFlag::BlockRead))		shareMode |= FILE_SHARE_READ;
	if (!(open->m_flags & SFS_OpenFlag::BlockWrite))	shareMode |= FILE_SHARE_WRITE;
	if (!(open->m_flags & SFS_OpenFlag::BlockDelete))	shareMode |= FILE_SHARE_DELETE;
			
	DWORD flagsAndAttrs = FILE_FLAG_OVERLAPPED;
	if (open->m_flags & SFS_OpenFlag::DeleteOnClose)
		flagsAndAttrs |= FILE_FLAG_DELETE_ON_CLOSE;

	if (sfsAttrs->m_validAttrFlags & SFS_Attr::Bits)
	{
		flagsAndAttrs |= ModifyBasicWinAttrBits(0, sfsAttrs);
		if (sfsAttrs->m_attrBitsValid & sfsAttrs->m_attrBits & SFS_AttrBit::Encrypted)
			flagsAndAttrs |= FILE_ATTRIBUTE_ENCRYPTED;
	}

	file->m_fileAccess = ToWinAccessRights(open->m_desiredAccess);
	if (open->m_flags & SFS_OpenFlag::AccessAuditAlarmInfo)
		file->m_fileAccess |= ACCESS_SYSTEM_SECURITY;
	if (creatDisp == TRUNCATE_EXISTING)
		file->m_fileAccess |= GENERIC_WRITE;

	file->m_fileHandle = CreateFile(file->m_winPath.c_str(), file->m_fileAccess, shareMode, 0, creatDisp, flagsAndAttrs, 0);
	if (!file->ValidFileHandle())
	{
		DWORD winError = GetLastError();
		unsigned int statusCode = ToSfsStatusCode(winError);

		SfsSendLog(Log(Log::Failure, 5 + 5)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"flags", open->m_flags)
						   .AddNum(L"desiredAccess", open->m_desiredAccess)
						   .AddStatusCode(statusCode)
			.AddGrp_Error_WinApi(L"CreateFile()", false, winError));

		throw SfsStatusResponse(statusCode);
	}
	
	bool createdNewFile = creatDisp == CREATE_NEW || 
		((creatDisp == CREATE_ALWAYS || creatDisp == OPEN_ALWAYS) && GetLastError() != ERROR_ALREADY_EXISTS);

	if (createdNewFile)
		file->m_transferFile.RegisterCreateNewFile(); // new file has been created
	else if (creatDisp == CREATE_ALWAYS || creatDisp == TRUNCATE_EXISTING)
		file->m_transferFile.RegisterResizeFile(); // existing file has been truncated
	
	bool partialFailure = false;
	if (createdNewFile)
	{
		if (winTimes.AnyPresent())
		{
			try
			{
				SetTimesT(file->m_fileHandle, winTimes);
			}
			catch (WinApiError const& e)
			{
				wchar_t const* message = L"Setting file time for the newly created file failed.";
				
				SfsSendLog(Log(SFS_LogAction::Open + SetFileTimeFailureAsProgress, message, false, 2 + 5)
						.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
					    .AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode()));

				partialFailure = true;
			}	
		}
	
		// At this point, the built-in FlowSfsWin would also process
		// the following attribute bits: SFS_AttrBit::Compressed,
		// SFS_AttrBit::Encrypted, and SFS_AttrBit::Sparse.
	}

	AddFsObj(file);
	autoFile.release();
				
	SfsSendLog(Log(Log::Success,6)
		.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
					   .AddNum(L"flags", open->m_flags)
					   .AddNum(L"desiredAccess", open->m_desiredAccess)
					   .AddBln(L"createdNewFile", createdNewFile)
					   .AddBln(L"partialFailure", partialFailure));

	SfsSendResponse(SfsHandleResponse(file->m_sfsHandle, createdNewFile));
}

void SfsWin::HandleClose(SFS_Request const* request)
{
	FsObj* obj = ExtractFsObj(request);

	if (obj->Root())
	{
		FsRoot* root = obj->Root();
		if (root->m_listDir.RegisterCount())
			SfsSendLogByHandle(obj->m_sfsHandle, Log(obj->m_winPath.c_str(), false, root->m_listDir));
	}
	else if (obj->Dir())
	{
		FsDir* dir = obj->Dir();
		dir->CloseFindHandle();

		if (dir->m_listDir.RegisterCount())
			SfsSendLogByHandle(obj->m_sfsHandle, Log(obj->m_winPath.c_str(), false, dir->m_listDir));
	}
	else // if (obj->File())
	{
		FsFile* file = obj->File();
		file->CloseFileHandle();

		if (file->m_transferFile.RegisterCount())
			SfsSendLogByHandle(obj->m_sfsHandle, Log(obj->m_winPath.c_str(), false, file->m_transferFile));
	}
	
	SfsSendLog(Log(Log::Success, 3)
		.AddGrp_Param().AddStr(L"path", obj->m_winPath.c_str())
					   .AddStr(L"type", obj->File() ? L"File" : L"Directory"));

	DelFsObj(obj->m_sfsHandle); // invalidates the obj pointer

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleRead(SFS_Request const* request)
{
	SFS_Read const* read = (SFS_Read const*) request->m_content;

	FsFile* file = ExtractFsFile(request);

	// In text mode, files are always read sequentially; the offset in
	// SFS_Read is ignored. SFS_TextSeek may be used to move the text
	// offset to a particular line. For more info, see the SFTPv6 draft.

	unsigned __int64 offset;
	if (file->m_textMode)
		offset = file->m_textOffset;
	else
		offset = read->m_offset;
	
	// SFS providers must be able to read data chunks up to 4 MB.
	unsigned int length = min(s_maxReadSize, read->m_length);
	
	AutoArray<byte> data(new byte[length]); 
	bool eofReached;
			
	try
	{
		eofReached = ReadT(file->m_fileHandle, offset, data.Get(), length);
	}
	catch (WinApiError const& e)
	{
		unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
		
		SfsSendLog(Log(Log::Failure, 5 + 5)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"offset", offset)
						   .AddNum(L"length", length)
						   .AddStatusCode(statusCode)
		    .AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode()));

		throw SfsStatusResponse(statusCode);
	}

	if (eofReached && !length)
	{
		if (!request->m_internal)
			file->m_transferFile.RegisterReadEof();

		SfsSendLog(Log(Log::Success, 6)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"offset", offset)
						   .AddNum(L"length", 0)
						   .AddBln(L"reachedEof", true)
						   .AddStatusCode(SFS_StatusCode::Eof));
		
		throw SfsStatusResponse(SFS_StatusCode::Eof);
	}
	else
	{
		if (file->m_textMode)
			file->m_textOffset += length;
			
		if (!request->m_internal)
			file->m_transferFile.RegisterRead(offset, length);

		SfsSendLog(Log(Log::Success, 5)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"offset", offset)
						   .AddNum(L"length", length)
						   .AddBln(L"reachedEof", eofReached));

		SFS_Data sfsData;
		sfsData.m_dataSize = length;
		sfsData.m_dataPtr = data.Get();
		sfsData.m_endOfFile = eofReached;

		SfsSendResponse(SfsResponse(SFS_ResponseType::Data, &sfsData));
	}
}
	
void SfsWin::HandleWrite(SFS_Request const* request)
{
	SFS_Write const* write = (SFS_Write const*) request->m_content;

	FsFile* file = ExtractFsFile(request);

	// In text mode, files are always written sequentially; the offset in
	// SFS_Write is ignored. SFS_TextSeek may be used to move the text
	// offset to a particular line. For more info, see the SFTPv6 draft.

	unsigned __int64 offset;
	if (file->m_append)
		offset = -1;
	else if (file->m_textMode)
		offset = file->m_textOffset;
	else
		offset = write->m_offset;

	bool const append = offset == -1;
	
	try
	{
		WriteT(file->m_fileHandle, offset, write->m_dataPtr, write->m_dataSize);
	}
	catch (WinApiError const& e)
	{
		unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
		
		Log log(Log::Failure, 5 + 5);
		log.AddGrp_Param();
		log.AddStr(L"path", file->m_winPath.c_str());
		if (append)
			log.AddStr(L"offset", L"Eof");
		else		
			log.AddNum(L"offset", offset);
		log.AddNum(L"length", write->m_dataSize);
		log.AddStatusCode(statusCode);
		log.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode());
		SfsSendLog(log);
		
		throw SfsStatusResponse(statusCode);
	}

	if (file->m_textMode && !append)
		file->m_textOffset += write->m_dataSize;

	if (!request->m_internal)
		file->m_transferFile.RegisterWrite(offset, write->m_dataSize);

	Log log(Log::Success, 4);
	log.AddGrp_Param();
	log.AddStr(L"path", file->m_winPath.c_str());
	if (append)
		log.AddStr(L"offset", L"Eof");
	else
		log.AddNum(L"offset", offset);
	log.AddNum(L"length", write->m_dataSize);
	SfsSendLog(log);

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleStat(SFS_Request const* request)
{
	FsObj* obj = 0;
	wstring winPath;
	HANDLE handle = 0;
	DWORD handleAccess = 0;
	unsigned int sfsFlags;
	wchar_t const* how = 0;

	if (request->m_type == SFS_RequestType::FStat)
	{
		obj = ExtractFsObj(request);
		winPath = obj->m_winPath;
		if (obj->File())
		{
			handle = obj->File()->m_fileHandle;
			handleAccess = obj->File()->m_fileAccess;
		}
	
		sfsFlags = ((SFS_FStat const*) request->m_content)->m_flags;

		how = L"ByHandle";
	}
	else
	{
		ExtractWinPath(request, winPath);
		
		SFS_Stat const* stat = (SFS_Stat const*) request->m_content;
		sfsFlags = stat->m_flags;

		// This sample does not support symbolic links. That's
		// why the stat->m_followSymlink is ignored here.
	}

	SFS_Attrs sfsAttrs;
	ZeroMemory(&sfsAttrs, sizeof(sfsAttrs));

	if (winPath.empty()) 
	{
		// Make up attributes for all drives.

		sfsAttrs.m_type = SFS_Type::Directory;
		sfsAttrs.m_size = 0;
		sfsAttrs.m_validAttrFlags |= SFS_Attr::Size;
		sfsAttrs.m_linkCount = 1;
		sfsAttrs.m_validAttrFlags |= SFS_Attr::LinkCount;
	
		Log log(Log::Success, 3);
		log.AddGrp_Param();
		log.AddStr(L"path", L"Drives");
		if (how) 
			log.AddStr(L"how", how);
		SfsSendLog(log);
	}
	else
	{
		try
		{
			if (handleAccess & FILE_READ_ATTRIBUTES)
			{
				BY_HANDLE_FILE_INFORMATION fileInfo;
				GetFileInfoT(handle, fileInfo);
				ToSfsAttrs(fileInfo, sfsAttrs);
			}
			else
			{
				try
				{
					BY_HANDLE_FILE_INFORMATION fileInfo;
					GetFileInfoT(winPath.c_str(), fileInfo);
					
					ToSfsAttrs(fileInfo, sfsAttrs);
				}
				catch (WinApiError const&)
				{
					WIN32_FILE_ATTRIBUTE_DATA fileInfo;
					GetFileInfoT(winPath.c_str(), fileInfo);

					ToSfsAttrs(fileInfo, sfsAttrs);
				}
			}
		}
		catch (WinApiError const& e)
		{
			unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
			
			Log log(Log::Failure, 4 + 5);
			log.AddGrp_Param();
			log.AddStr(L"path", winPath.c_str());
			if (how) 
				log.AddStr(L"how", how);
			log.AddStatusCode(statusCode);
			log.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode());
			SfsSendLog(log);

			throw SfsStatusResponse(statusCode);
		}

		// If requested via sfsFlags, the built-in FlowSshWin would
		// also retrieve SFS_Attr::OwnerGroup, SFS_Attr::Permissions,
		// SFS_Attr::Acl, and SFS_Attr::TextHint.

		Log log(Log::Success, 3);
		log.AddGrp_Param();
		log.AddStr(L"path", winPath.c_str());
		if (how)
			log.AddStr(L"how", how);
		SfsSendLog(log);
	}

	SfsSendResponse(SfsResponse(SFS_ResponseType::Attrs, &sfsAttrs));
}

void SfsWin::HandleSetStat(SFS_Request const* request)
{
	SFS_Attrs const* sfsAttrs = ExtractSfsAttrs(request);

	FsObj* obj = 0;
	wstring winPath;
	HANDLE handle = 0;
	DWORD handleAccess = 0;
	wchar_t const* how = 0;
	bool partialSuccess = false;

	if (request->m_type == SFS_RequestType::SetStat)
		ExtractWinPath(request, winPath);
	else
	{
		obj = ExtractFsObj(request);
		winPath = obj->m_winPath;
		FsFile* file = obj->File();
		if (file)
		{
			handle = file->m_fileHandle;
			handleAccess = file->m_fileAccess;
		}

		how = L"ByHandle";
	}

	if (winPath.empty())
	{
		Log log(Log::Failure, 5 + 3);
		log.AddGrp_Param();
		log.AddStr(L"path", L"Drives");
		if (how)
			log.AddStr(L"how", how);
		log.AddStatusCode(SFS_StatusCode::PermissionDenied);
		log.AddGrp_Error_Exception(L"Access to logical drives is limited to read-only.", false);
		SfsSendLog(log);

		throw SfsStatusResponse(SFS_StatusCode::PermissionDenied);
	}

	WinTimes winTimes;
	ExtractWinTimes(request, winTimes);

	if (sfsAttrs->m_validAttrFlags & (SFS_Attr::Permissions | SFS_Attr::Acl | SFS_Attr::OwnerGroup | SFS_Attr::AllocSize | SFS_Attr::CTime))
	{
		wchar_t const* reason = L"The following attributes are not supported: Permissions, Acl, OwnerGroup, AllocSize, CTime.";
		
		Log log(Log::Failure, 5 + 3);
		log.AddGrp_Param();
		log.AddStr(L"path", winPath.c_str());
		if (how)
			log.AddStr(L"how", how);
		log.AddBln(L"partialSuccess", partialSuccess);
		log.AddStatusCode(SFS_StatusCode::Failure);
		log.AddGrp_Error_Exception(reason, false);
		SfsSendLog(log);

		throw SfsStatusResponse(SFS_StatusCode::Failure, reason);
	}
	
	// Open or verify existing handle.
	AutoHandle autoHandle;
	bool readAttrs = false;
	DWORD winAttrs = INVALID_FILE_ATTRIBUTES;

	try
	{
		DWORD requiredAccess = 0;

		if (sfsAttrs->m_validAttrFlags & SFS_Attr::Size)
			requiredAccess |= FILE_WRITE_DATA;
		if (sfsAttrs->m_validAttrFlags & (SFS_Attr::AccessTime | SFS_Attr::CreateTime | SFS_Attr::ModifyTime))
			requiredAccess |= FILE_WRITE_ATTRIBUTES;
		if (sfsAttrs->m_validAttrFlags & SFS_Attr::Bits)
			readAttrs = true;
		
		if (readAttrs || !handle)
		{
			if (handleAccess & FILE_READ_ATTRIBUTES)
			{
				BY_HANDLE_FILE_INFORMATION fileInfo;
				GetFileInfoT(handle, fileInfo);
				winAttrs = fileInfo.dwFileAttributes;
			}
			else
			{
				winAttrs = GetFileAttributes(winPath.c_str());
				if (winAttrs == INVALID_FILE_ATTRIBUTES)
					throw WinApiError(L"GetFileAttributes()");
			}
		}
		
		if (!handle || (handleAccess & requiredAccess) != requiredAccess)
		{
			try
			{
				DWORD openFlags = 0;
				if (!handle && (winAttrs & FILE_ATTRIBUTE_DIRECTORY))
					openFlags |= FILE_FLAG_BACKUP_SEMANTICS;

				autoHandle.Set(OpenExistingT(winPath.c_str(), requiredAccess, openFlags));
				handle = autoHandle.Get();
				handleAccess = requiredAccess;
			}
			catch (WinApiError const& e)
			{
				// Throw if action cannot be perform without a handle.

				if ((e.GetErrorCode() != ERROR_ACCESS_DENIED) ||
					(sfsAttrs->m_validAttrFlags & SFS_Attr::Size) ||
					(sfsAttrs->m_validAttrFlags & (SFS_Attr::AccessTime | SFS_Attr::CreateTime | SFS_Attr::ModifyTime)))
					throw;
			}
		}

		if (sfsAttrs->m_validAttrFlags & SFS_Attr::Size)
		{
			// Note that handle can't be invalid at this point.
			SetSizeT(handle, sfsAttrs->m_size);
			partialSuccess = true;

			if (obj)
			{
				if (obj->File())
					obj->File()->m_transferFile.RegisterResizeFile();
			}
			else
			{
				SFS::TransferFile transferFile;
				transferFile.RegisterResizeFile();
				SfsSendLog(Log(winPath.c_str(), false, transferFile));
			}
		}

		if (sfsAttrs->m_validAttrFlags & SFS_Attr::Bits)
		{
			DWORD newWinAttrs = RemoveReadOnlyWinAttrBits(ModifyBasicWinAttrBits(winAttrs, sfsAttrs));
			if (newWinAttrs != RemoveReadOnlyWinAttrBits(winAttrs))
			{
				if (!SetFileAttributes(winPath.c_str(), newWinAttrs))
					throw WinApiError(L"SetFileAttributes()");
				partialSuccess = true;
			}

			// At this point, the built-in FlowSfsWin would also process
			// the following attribute bits: SFS_AttrBit::Compressed,
			// SFS_AttrBit::Encrypted, and SFS_AttrBit::Sparse.
		}
		
		if (winTimes.AnyPresent())
		{
			// Note that handle can't be invalid at this point.
			SetTimesT(handle, winTimes);
			partialSuccess = true;
		}
	}
	catch (WinApiError const& e)
	{	
		unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
	
		Log log(Log::Failure, 5 + 5);
		log.AddGrp_Param();
		log.AddStr(L"path", winPath.c_str());
		if (how)
			log.AddStr(L"how", how);
		log.AddBln(L"partialSuccess", partialSuccess);
		log.AddStatusCode(statusCode);
		log.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode());
		SfsSendLog(log);

		throw SfsStatusResponse(statusCode);
	}
	
	Log log(Log::Success, 3);
	log.AddGrp_Param();
	log.AddStr(L"path", winPath.c_str());
	if (how)
		log.AddStr(L"how", how);
	SfsSendLog(log);

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleOpenDir(SFS_Request const* request)
{
	SFS_OpenDir const* openDir = (SFS_OpenDir const*) request->m_content;

	wstring winPath;
	ExtractWinPath(request, winPath);

	if (winPath.empty())
	{ 
		auto_ptr<FsRoot> autoRoot(new FsRoot);
		FsRoot* root = autoRoot.get();
		root->m_sfsPath = openDir->m_path;
		root->m_winPath = winPath;
			
		root->m_drives = GetLogicalDrives();
		if (!root->m_drives)
		{
			DWORD winError = GetLastError();
			unsigned int statusCode = ToSfsStatusCode(winError);

			SfsSendLog(Log(Log::Failure, 3 + 5)
				.AddGrp_Param().AddStr(L"path", L"Drives")
							   .AddStatusCode(statusCode)
				.AddGrp_Error_WinApi(L"GetLogicalDrives()", false, winError));

			throw SfsStatusResponse(statusCode);
		}

		AddFsObj(root);
		autoRoot.release();

		SfsSendLog(Log(Log::Success, 2)
			.AddGrp_Param().AddStr(L"path", L"Drives"));

		SfsSendResponse(SfsHandleResponse(root->m_sfsHandle));
	}
	else
	{
		auto_ptr<FsDir> autoDir(new FsDir);
		FsDir* dir = autoDir.get();
		dir->m_sfsPath = openDir->m_path;
		dir->m_winPath = winPath;
	
		wstring findName;
		findName.reserve(winPath.size() + 2);
		findName = winPath;
		if (winPath.size() && winPath[winPath.size()-1] != L'\\')
			findName += L"\\*";
		else
			findName += L"*";
		
		dir->m_findHandle = FindFirstFile(findName.c_str(), &dir->m_findDelayed);
		if (!dir->ValidFindHandle() && GetLastError() != ERROR_FILE_NOT_FOUND)
		{
			DWORD winError = GetLastError();
			unsigned int statusCode = ToSfsStatusCode(winError);
			if (IsFile(winPath.c_str()))
				statusCode = SFS_StatusCode::NotADirectory;

			SfsSendLog(Log(Log::Failure, 3 + 5)
				.AddGrp_Param().AddStr(L"path", winPath.c_str())
							   .AddStatusCode(statusCode)
				.AddGrp_Error_WinApi(L"FindFirstFile()", false, winError));

			throw SfsStatusResponse(statusCode);
		}
		
		if (!dir->ValidFindHandle())
			dir->m_noMoreFiles = true;
		else if (FindFileFilter(dir->m_findDelayed))
			dir->m_findDelayedPresent = true;
		
		AddFsObj(dir);
		autoDir.release();

		SfsSendLog(Log(Log::Success, 2)
			.AddGrp_Param().AddStr(L"path", winPath.c_str()));

		SfsSendResponse(SfsHandleResponse(dir->m_sfsHandle));
	}
}

void SfsWin::HandleReadDir(SFS_Request const* request)
{
	FsObj* obj = ExtractFsRootOrDir(request);
	SFS_ReadDir const* readDir = (SFS_ReadDir const*) request->m_content;

	if (obj->Root())
	{
		FsRoot* root = obj->Root();
		if (!root->m_drives)
		{
			if (!request->m_internal)
				root->m_listDir.RegisterListEof();

			SfsSendLog(Log(Log::Success, 5)
				.AddGrp_Param().AddStr(L"path", L"Drives")
							   .AddNum(L"entriesCount", 0)
							   .AddBln(L"reachedEof", true)
							   .AddStatusCode(SFS_StatusCode::Eof));
			
			throw SfsStatusResponse(SFS_StatusCode::Eof);
		}
		
		AutoArray<SFS_Name> names(new SFS_Name[33]);
		ZeroMemory(names.Get(), sizeof(SFS_Name) * 33);
		vector<wstring> strings;
		strings.resize(33);

		unsigned int count = 0;
		
		strings[count] = L".";
		names[count].m_fileName = strings[count].c_str();
		names[count].m_attrs.m_type = SFS_Type::Directory;
		names[count].m_attrs.m_validAttrFlags = SFS_Attr::Size;
		++count;

		if (m_mountPath != L"/") // if not mounted on the root of SFS virtual path
		{
			strings[count] = L"..";
			names[count].m_fileName = strings[count].c_str();
			names[count].m_attrs.m_type = SFS_Type::Directory;
			names[count].m_attrs.m_validAttrFlags = SFS_Attr::Size;
			++count;
		}
					
		for ( ; root->m_drives; root->m_drives >>= 1, ++root->m_nextDrive)
		{
			if (root->m_drives & 1)
			{
				wchar_t path[4] = { root->m_nextDrive, L':', L'\\', L'\0' };
				strings[count].assign(path, 1);
				names[count].m_fileName = strings[count].c_str();
				
				try
				{
					BY_HANDLE_FILE_INFORMATION fileInfo;
					GetFileInfoT(path, fileInfo);
					ToSfsAttrs(fileInfo, names[count].m_attrs);
				}
				catch (WinApiError const&)
				{
					names[count].m_attrs.m_type = SFS_Type::Directory;
					names[count].m_attrs.m_validAttrFlags = SFS_Attr::Size;
				}

				++count;
			}
		}

		if (!request->m_internal)
			root->m_listDir.RegisterList(count);

		SfsSendLog(Log(Log::Success, 4)
			.AddGrp_Param().AddStr(L"path", L"Drives")
						   .AddNum(L"entriesCount", count)
						   .AddBln(L"reachedEof", true));
		
		SFS_Names sfsNames;
		sfsNames.m_nameCount = count;
		sfsNames.m_nameArray = names.Get();
		sfsNames.m_endOfList = true;
		SfsSendResponse(SfsResponse(SFS_ResponseType::Names, &sfsNames));
	}
	else // if (obj->Dir())
	{
		FsDir* dir = obj->Dir();
		if (dir->m_noMoreFiles && !dir->m_findDelayedPresent)
		{
			if (!request->m_internal)
				dir->m_listDir.RegisterListEof();

			SfsSendLog(Log(Log::Success, 5)
				.AddGrp_Param().AddStr(L"path", dir->m_winPath.c_str())
							   .AddNum(L"entriesCount", 0)
							   .AddBln(L"reachedEof", true)
							   .AddStatusCode(SFS_StatusCode::Eof));
			
			throw SfsStatusResponse(SFS_StatusCode::Eof);
		}
		
		unsigned int const maxCount = 50;
		unsigned int count = 0;

		AutoArray<SFS_Name> names(new SFS_Name[maxCount]);
		ZeroMemory(names.Get(), sizeof(SFS_Name) * maxCount);
		vector<wstring> strings;
		strings.resize(maxCount);

		if (dir->m_firstReadDir)
		{	
			strings[count] = L".";
			names[count].m_fileName = strings[count].c_str();
			names[count].m_attrs.m_type = SFS_Type::Directory;
			names[count].m_attrs.m_validAttrFlags = SFS_Attr::Size;
			++count;

			strings[count] = L"..";
			names[count].m_fileName = strings[count].c_str();
			names[count].m_attrs.m_type = SFS_Type::Directory;
			names[count].m_attrs.m_validAttrFlags = SFS_Attr::Size;
			++count;
		}
		
		if (dir->m_findDelayedPresent)
		{
			strings[count] = dir->m_findDelayed.cFileName;
			names[count].m_fileName = strings[count].c_str();
			ToSfsAttrs(dir->m_findDelayed, names[count].m_attrs);
			++count;

			dir->m_findDelayedPresent = false;
		}

		if (!dir->m_noMoreFiles)
		{
			while (!dir->m_noMoreFiles)
			{
				if (!FindNextFile(dir->m_findHandle, &dir->m_findDelayed))
				{
					DWORD winError = GetLastError();
					if (winError == ERROR_NO_MORE_FILES)
						dir->m_noMoreFiles = true;
					
					if (count)
					{
						// Ignore this error if we have at least one item to send.
						// The error will be reported in the next ReadDir attempt.
						break; 
					}
					else if (dir->m_noMoreFiles)
					{
						if (!request->m_internal)
							dir->m_listDir.RegisterListEof();

						SfsSendLog(Log(Log::Success, 5)
							.AddGrp_Param().AddStr(L"path", dir->m_winPath.c_str())
										   .AddNum(L"entriesCount", 0)
										   .AddBln(L"reachedEof", true)
										   .AddStatusCode(SFS_StatusCode::Eof));
					
						throw SfsStatusResponse(SFS_StatusCode::Eof);
					}
					else
					{
						unsigned int statusCode = ToSfsStatusCode(winError);

						SfsSendLog(Log(Log::Failure, 3 + 5)
							.AddGrp_Param().AddStr(L"path", dir->m_winPath.c_str())
										   .AddStatusCode(statusCode)
							.AddGrp_Error_WinApi(L"FindNextFile()", false, winError));

						throw SfsStatusResponse(statusCode);
					}
				}
				else if (FindFileFilter(dir->m_findDelayed))
				{
					if (count == maxCount)
					{
						dir->m_findDelayedPresent = true;
						break;
					}

					strings[count] = dir->m_findDelayed.cFileName;
					names[count].m_fileName = strings[count].c_str();
					ToSfsAttrs(dir->m_findDelayed, names[count].m_attrs);
					++count;
				}
			}
		}
			
		dir->m_firstReadDir = false;

		if (!request->m_internal)
			dir->m_listDir.RegisterList(count);

		bool endOfList = dir->m_noMoreFiles && !dir->m_findDelayedPresent;

		SfsSendLog(Log(Log::Success, 4)
			.AddGrp_Param().AddStr(L"path", dir->m_winPath.c_str())
						   .AddNum(L"entriesCount", count)
						   .AddBln(L"reachedEof", endOfList));
		
		SFS_Names sfsNames;
		sfsNames.m_nameCount = count;
		sfsNames.m_nameArray = names.Get();
		sfsNames.m_endOfList = endOfList;
		SfsSendResponse(SfsResponse(SFS_ResponseType::Names, &sfsNames));
	}
}

void SfsWin::HandleRemove(SFS_Request const* request)
{
	wstring winPath;
	ExtractWinPath(request, winPath);
	
	if (!DeleteFile(winPath.c_str()))
	{
		DWORD winError = GetLastError();
		unsigned int statusCode = ToSfsStatusCode(winError);
		if (IsDirectory(winPath.c_str()))
			statusCode = SFS_StatusCode::FileIsADirectory;
			
		SfsSendLog(Log(Log::Failure, 3 + 5)
			.AddGrp_Param().AddStr(L"path", winPath.c_str())
						   .AddStatusCode(statusCode)
			.AddGrp_Error_WinApi(L"DeleteFile()", false, winError));

		throw SfsStatusResponse(statusCode);
	}

	SfsSendLog(Log(Log::Success, 2)
		.AddGrp_Param().AddStr(L"path", winPath.c_str()));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleMkDir(SFS_Request const* request)
{
	wstring winPath;
	ExtractWinPath(request, winPath);

	SFS_Attrs const* sfsAttrs = ExtractSfsAttrs(request);
	
	WinTimes winTimes;
	ExtractWinTimes(request, winTimes);

	// At this point, the built-in FlowSfsWin would also extract security
	// attributes (owner, group, POSIX permissions, ACL) which it would
	// then pass on to the CreateDirectory() call.
	
	if (!CreateDirectory(winPath.c_str(), 0))
	{
		// Note that CreateDirectory() will create the top level directory only,
		// failing if the parent directory does not yet exist. On the other hand, 
		// the built-in FlowSfsWin is able to create a multi-level directory here.

		DWORD winError = GetLastError();
		unsigned int statusCode = ToSfsStatusCode(winError);

		SfsSendLog(Log(Log::Failure, 3 + 5)
			.AddGrp_Param().AddStr(L"path", winPath.c_str())
						   .AddStatusCode(statusCode)
			.AddGrp_Error_WinApi(L"CreateDirectory()", false, winError));

		throw SfsStatusResponse(statusCode);
	}

	bool partialFailure = false;

	if (winTimes.AnyPresent())
	{
		try
		{
			AutoHandle handle(OpenExistingT(winPath.c_str(), FILE_WRITE_ATTRIBUTES, FILE_FLAG_BACKUP_SEMANTICS));
			SetTimesT(handle, winTimes);
		}
		catch (WinApiError const& e)
		{
			wchar_t const* message = L"Setting file time for the newly created directory failed.";

			SfsSendLog(Log(SFS_LogAction::MkDir + SetFileTimeFailureAsProgress, message, false, 2 + 5)
				.AddGrp_Param().AddStr(L"path", winPath.c_str())
				.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode()));
			
			partialFailure = true;
		}
	}

	if (sfsAttrs->m_validAttrFlags & SFS_Attr::Bits)
	{
		DWORD winAttrs = GetFileAttributes(winPath.c_str());
		if (winAttrs == INVALID_FILE_ATTRIBUTES)
		{
			wchar_t const* message = L"Setting file attributes for the newly created directory failed.";
			
			DWORD winError = GetLastError();
		
			SfsSendLog(Log(SFS_LogAction::MkDir + SetFileAttrsFailureAsProgress, message, false, 2 + 5)
				.AddGrp_Param().AddStr(L"path", winPath.c_str())
				.AddGrp_Error_WinApi(L"GetFileAttributes()", false, winError));
			
			partialFailure = true;
		}
		else
		{
			DWORD newWinAttrs = RemoveReadOnlyWinAttrBits(ModifyBasicWinAttrBits(winAttrs, sfsAttrs));
			if (newWinAttrs != RemoveReadOnlyWinAttrBits(winAttrs))
			{
				if (!SetFileAttributes(winPath.c_str(), newWinAttrs))
				{
					DWORD winError = GetLastError();

					wchar_t const* message = L"Setting file attributes for the newly created directory failed.";

					SfsSendLog(Log(SFS_LogAction::MkDir + SetFileAttrsFailureAsProgress, message, false, 2 + 5)
						.AddGrp_Param().AddStr(L"path", winPath.c_str())
						.AddGrp_Error_WinApi(L"SetFileAttributes()", false, winError));

					partialFailure = true;
				}

				// At this point, the built-in FlowSfsWin would also process
				// the following attribute bits: SFS_AttrBit::Compressed,
				// and SFS_AttrBit::Encrypted
			}
		}
	}

	SfsSendLog(Log(Log::Success, 3)
		.AddGrp_Param().AddStr(L"path", winPath.c_str())
					   .AddBln(L"partialFailure", partialFailure));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleRmDir(SFS_Request const* request)
{
	wstring winPath;
	ExtractWinPath(request, winPath);

	if (!RemoveDirectory(winPath.c_str()))
	{
		DWORD winError = GetLastError();
		unsigned int statusCode = ToSfsStatusCode(winError);
		if (IsFile(winPath.c_str()))
			statusCode = SFS_StatusCode::NotADirectory;
		
		SfsSendLog(Log(Log::Failure, 3 + 5)
			.AddGrp_Param().AddStr(L"path", winPath.c_str())
						   .AddStatusCode(statusCode)
		    .AddGrp_Error_WinApi(L"RemoveDirectory()", false, winError));

		throw SfsStatusResponse(statusCode);
	}

	SfsSendLog(Log(Log::Success, 3 + 5)
			.AddGrp_Param().AddStr(L"path", winPath.c_str()));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleRealPath(SFS_Request const* request)
{
	SFS_RealPath const* realPath = (SFS_RealPath const*) request->m_content;

	wstring winPath;
	SfsToWinPath(realPath->m_path, winPath);

	SFS_Name name;
	ZeroMemory(&name, sizeof(name));
	name.m_fileName = realPath->m_path;
	name.m_attrs.m_type = SFS_Type::Unknown;
	
	if (realPath->m_controlByte != SFS_ControlByte::NoCheck)
	{
		try
		{
			BY_HANDLE_FILE_INFORMATION fileInfo;
			GetFileInfoT(winPath.c_str(), fileInfo);
			ToSfsAttrs(fileInfo, name.m_attrs);
		}
		catch (WinApiError const& e)
		{
			if (realPath->m_controlByte == SFS_ControlByte::StatAlways)
			{
				unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
				
				SfsSendLog(Log(Log::Failure, 5 + 5)
					.AddGrp_Param().AddStr(L"path", winPath.c_str())
								   .AddControlByte(realPath->m_controlByte)
								   .AddStr(L"how", L"WithStat")
								   .AddStatusCode(statusCode)
					.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode()));

				throw SfsStatusResponse(statusCode);
			}
		}
	}

	Log log(Log::Success, 4);
	log.AddGrp_Param();
	log.AddStr(L"path", winPath.c_str());
	log.AddControlByte(realPath->m_controlByte);
	if (name.m_attrs.m_type != SFS_Type::Unknown)
		log.AddStr(L"how", L"WithStat");
	SfsSendLog(log);
		
	SfsSendResponse(SfsResponse(SFS_ResponseType::Name, &name));
}

void SfsWin::HandleRename(SFS_Request const* request)
{
	wstring oldWinPath, newWinPath;
	ExtractWinPaths(request, oldWinPath, newWinPath);

	unsigned int sfsFlags = ((SFS_Rename const*) request->m_content)->m_flags;

	DWORD winFlags = 0;
	if (sfsFlags & SFS_RenameFlag::Overwrite)
		winFlags |= MOVEFILE_REPLACE_EXISTING;
	if ((sfsFlags & (SFS_RenameFlag::Atomic | SFS_RenameFlag::Native)) != SFS_RenameFlag::Atomic)
		winFlags |= MOVEFILE_COPY_ALLOWED;

	if (!MoveFileEx(oldWinPath.c_str(), newWinPath.c_str(), winFlags))
	{
		DWORD winError = GetLastError();
		unsigned int statusCode = ToSfsStatusCode(winError);
		
		SfsSendLog(Log(Log::Failure, 5 + 5)
			.AddGrp_Param().AddStr(L"oldPath", oldWinPath.c_str())
						   .AddStr(L"newPath", newWinPath.c_str())
						   .AddNum(L"flags", sfsFlags)
						   .AddStatusCode(statusCode)
			.AddGrp_Error_WinApi(L"MoveFileEx()", false, winError));

		throw SfsStatusResponse(statusCode);
	}

	SfsSendLog(Log(Log::Success, 4)
		.AddGrp_Param().AddStr(L"oldPath", oldWinPath.c_str())
					   .AddStr(L"newPath", newWinPath.c_str())
					   .AddNum(L"flags", sfsFlags));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleReadLink(SFS_Request const* request)
{
	wstring winPath;
	ExtractWinPath(request, winPath);

	// The built-in FlowSfsWin does support symbolic links.

	SfsSendLog(Log(Log::Failure, 3 + 3)
		.AddGrp_Param().AddStr(L"path", winPath.c_str())
					   .AddStatusCode(SFS_StatusCode::OpUnsupported)
		.AddGrp_Error_Exception(L"Symbolic links are not supported by this provider.", false));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::OpUnsupported, L"The operation is not supported on the provided path."));
}

void SfsWin::HandleLink(SFS_Request const* request)
{
	SFS_Link const* link = (SFS_Link const*) request->m_content;

	wstring newLinkWinPath, existingWinPath;
	ExtractWinPaths(request, newLinkWinPath, existingWinPath);
	
	// The built-in FlowSfsWin does support symbolic and hard links.

	SfsSendLog(Log(Log::Failure, 5 + 3)
		.AddGrp_Param().AddStr(L"newLinkPath", newLinkWinPath.c_str())
					   .AddStr(L"existingPath", existingWinPath.c_str())
					   .AddStr(L"linkType", link->m_symlink ? L"Symbolic" : L"Hard")
					   .AddStatusCode(SFS_StatusCode::OpUnsupported)
		.AddGrp_Error_Exception(L"Symbolic and hard links are not supported by this provider.", false));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::OpUnsupported, L"The operation is not supported on the provided path."));
}

void SfsWin::HandleBlock(SFS_Request const* request)
{
	SFS_Block const* block = (SFS_Block const*) request->m_content;

	FsFile* file = ExtractFsFile(request);
	
	if (!(block->m_flags & (SFS_OpenFlag::BlockRead | SFS_OpenFlag::BlockWrite)))
	{
		wchar_t const* reason = L"The specified block flags are invalid.";

		SfsSendLog(Log(Log::Failure, 6 + 3)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"offset", block->m_offset)
						   .AddNum(L"length", block->m_length)
						   .AddNum(L"flags", block->m_flags)
						   .AddStatusCode(SFS_StatusCode::InvalidParameter)
			.AddGrp_Error_Exception(reason, false));

		throw SfsStatusResponse(SFS_StatusCode::InvalidParameter, reason);
	}
	
	ULARGE_INTEGER i;
	i.QuadPart = block->m_offset;
	m_overlapped.Offset = i.LowPart;
	m_overlapped.OffsetHigh = i.HighPart;
	ResetEvent(m_overlapped.hEvent);
		
	i.QuadPart = block->m_length;
		
	DWORD flags = LOCKFILE_FAIL_IMMEDIATELY;
	if (block->m_flags & SFS_OpenFlag::BlockWrite)
		flags |= LOCKFILE_EXCLUSIVE_LOCK;

	try
	{
		if (!LockFileEx(file->m_fileHandle, flags, 0, i.LowPart, i.HighPart, &m_overlapped))
		{
			if (GetLastError() != ERROR_IO_PENDING)
				throw WinApiError(L"LockFileEx()");

			while (true)
			{
				SfsAddWaitObject(m_overlapped.hEvent);
				SfsWait();

				DWORD dw;
				if (GetOverlappedResult(file->m_fileHandle, &m_overlapped, &dw, false))
					break;
				
				if (GetLastError() != ERROR_IO_INCOMPLETE)
					throw WinApiError(L"GetOverlappedResult()");
			}
		}
	}
	catch (WinApiError const& e)
	{
		unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
		if (e.GetErrorCode() == ERROR_LOCK_VIOLATION)
			statusCode = SFS_StatusCode::ByteRangeLockRefused;

		SfsSendLog(Log(Log::Failure, 6 + 5)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"offset", block->m_offset)
						   .AddNum(L"length", block->m_length)
						   .AddNum(L"flags", block->m_flags)
						   .AddStatusCode(statusCode)
			.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode()));

		throw SfsStatusResponse(statusCode);
	}

	SfsSendLog(Log(Log::Success, 5)
		.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
					   .AddNum(L"offset", block->m_offset)
					   .AddNum(L"length", block->m_length)
					   .AddNum(L"flags", block->m_flags));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleUnblock(SFS_Request const* request)
{
	SFS_Unblock const* unblock = (SFS_Unblock const*) request->m_content;

	FsFile* file = ExtractFsFile(request);
		
	ULARGE_INTEGER i;
	i.QuadPart = unblock->m_offset;
	m_overlapped.Offset = i.LowPart;
	m_overlapped.OffsetHigh = i.HighPart;
	ResetEvent(m_overlapped.hEvent);

		
	i.QuadPart = unblock->m_length;
		
	try
	{
		if (!UnlockFileEx(file->m_fileHandle, 0, i.LowPart, i.HighPart, &m_overlapped))
		{
			if (GetLastError() != ERROR_IO_PENDING)
				throw WinApiError(L"UnlockFileEx()");

			while (true)
			{
				SfsAddWaitObject(m_overlapped.hEvent);
				SfsWait();

				DWORD dw;
				if (GetOverlappedResult(file->m_fileHandle, &m_overlapped, &dw, false))
					break;
				
				if (GetLastError() != ERROR_IO_INCOMPLETE)
					throw WinApiError(L"GetOverlappedResult()");
			}
		}
	}
	catch (WinApiError const& e)
	{
		unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
		if (e.GetErrorCode() == ERROR_NOT_LOCKED)
			statusCode = SFS_StatusCode::NoMatchingByteRangeLock;

		SfsSendLog(Log(Log::Failure, 5 + 5)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"offset", unblock->m_offset)
						   .AddNum(L"length", unblock->m_length)
						   .AddStatusCode(statusCode)
			.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode()));

		throw SfsStatusResponse(statusCode);
	}

	SfsSendLog(Log(Log::Success, 4)
		.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
					   .AddNum(L"offset", unblock->m_offset)
					   .AddNum(L"length", unblock->m_length));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
}

void SfsWin::HandleTextSeek(SFS_Request const* request)
{
	SFS_TextSeek const* seek = (SFS_TextSeek const*) request->m_content;
	FsFile* file = ExtractFsFile(request);
	
	if (!file->m_textMode)
	{
		wchar_t const* reason = L"The file was not opened in text mode."; 

		SfsSendLog(Log(Log::Failure, 4 + 3)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"lineNumber", seek->m_lineNumber)
						   .AddStatusCode(SFS_StatusCode::InvalidHandle)
			.AddGrp_Error_Exception(reason, false));

		throw SfsStatusResponse(SFS_StatusCode::InvalidHandle, reason);
	}

	unsigned __int64 linesToGo = seek->m_lineNumber;
	if (!linesToGo)
	{
		file->m_textOffset = 0;

		SfsSendLog(Log(Log::Success, 4)
			.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
						   .AddNum(L"lineNumber", seek->m_lineNumber)
						   .AddNum(L"offset", 0));

		SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
	}
	else
	{
		unsigned __int64 offset = 0;
		unsigned int chunkSize = s_maxReadSize;
		{
			// Find best chunk size.
			ULARGE_INTEGER size;
			if (GetSize(file->m_fileHandle, size))
				for	( ; chunkSize > 1024; chunkSize >>= 1)
					if (size.QuadPart > (chunkSize >> 1))
						break;
		}

		byte lastByte = 0;
		bool eofReached;

		AutoArray<unsigned char> data(new unsigned char[chunkSize]);

		do
		{
			unsigned int dataSize = chunkSize;

			try
			{
				eofReached = ReadT(file->m_fileHandle, offset, data, dataSize);
			}
			catch (WinApiError const& e)
			{			
				unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());
				
				SfsSendLog(Log(Log::Failure, 4 + 5)
					.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
								   .AddNum(L"lineNumber", seek->m_lineNumber)
								   .AddStatusCode(statusCode)
					.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode()));

				throw SfsStatusResponse(statusCode);	
			}

			if (dataSize)
			{
				unsigned int i;
				for (i = 0; i < dataSize && linesToGo; ++i)
					if (10 == data[i])
						if (13 == (i ? data[i - 1] : lastByte))
							--linesToGo;
				
				offset += i;
				lastByte = data[dataSize - 1];
			}
		}
		while (!eofReached && linesToGo);
	
		file->m_textOffset = offset;

		if (!linesToGo)
		{
			SfsSendLog(Log(Log::Success, 4)
				.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
							   .AddNum(L"lineNumber", seek->m_lineNumber)
							   .AddNum(L"offset", offset));

		   SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Ok));
		}
		else
		{
			SfsSendLog(Log(Log::Success, 5)
				.AddGrp_Param().AddStr(L"path", file->m_winPath.c_str())
							   .AddNum(L"lineNumber", seek->m_lineNumber)
							   .AddNum(L"offset", offset)
							   .AddStatusCode(SFS_StatusCode::Eof));

		   SfsSendResponse(SfsStatusResponse(SFS_StatusCode::Eof));
		}
	}
}

void SfsWin::HandleCheckFile(SFS_Request const* request)
{
	wstring winPath;
	wstring hashAlg;
	
	AutoArray<unsigned char> hashData;
	unsigned int hashDataSize = 0;
	unsigned int hashDataCapacity = 0;
	
	wchar_t const* hashAlgList;
	unsigned __int64 startOffset;
	unsigned __int64 length;
	unsigned int blockSize;
	
	wchar_t const* how = 0;

	try
	{
		HANDLE fileHandle;
		AutoHandle autoFileHandle;
		
		if (request->m_type == SFS_RequestType::CheckFileHandle)
		{
			SFS_CheckFileHandle const* check = (SFS_CheckFileHandle const*) request->m_content;
			
			FsFile* file = ExtractFsFile(request);
			winPath = file->m_winPath;

			fileHandle = file->m_fileHandle;
			
			hashAlgList = check->m_hashAlgList;
			startOffset = check->m_startOffset;
			length = check->m_length;
			blockSize = check->m_blockSize;

			how = L"ByHandle";
		}
		else
		{
			SFS_CheckFileName const* check = (SFS_CheckFileName const*) request->m_content;
			
			ExtractWinPath(request, winPath);

			hashAlgList = check->m_hashAlgList;
			startOffset = check->m_startOffset;
			length = check->m_length;
			blockSize = check->m_blockSize;
			
			if (IsDirectory(winPath.c_str()))
			{
				SfsSendLog(Log(Log::Failure, 3 + 3)
					.AddGrp_Param().AddStr(L"path", winPath.c_str())
								   .AddStatusCode(SFS_StatusCode::FileIsADirectory)
					.AddGrp_Error_StatusCode(SFS_StatusCode::FileIsADirectory));

				throw SfsStatusResponse(SFS_StatusCode::FileIsADirectory);
			}
			
			DWORD access = FILE_READ_DATA;
			DWORD flags = FILE_FLAG_OVERLAPPED;

			try
			{
				// Try with FILE_READ_ATTRIBUTES first so that we can get the file size later on.
				autoFileHandle.Set(OpenExistingT(winPath.c_str(), access | FILE_READ_ATTRIBUTES, flags));
			}
			catch (WinApiError const&)
			{
				autoFileHandle.Set(OpenExistingT(winPath.c_str(), access, flags));
			}

			fileHandle = autoFileHandle.Get();
		}

		if (blockSize && blockSize < 256)
		{
			wchar_t const* reason = L"BlockSize must be least 256 bytes in size.";

			Log log(Log::Failure, 5 + 3);
			log.AddGrp_Param().AddStr(L"path", winPath.c_str());
			if (how) log.AddStr(L"how", how);
			log.AddNum(L"blockSize", blockSize)
			   .AddStatusCode(SFS_StatusCode::InvalidParameter);
			log.AddGrp_Error_Exception(reason, false);
			SfsSendLog(log);

			throw SfsStatusResponse(SFS_StatusCode::InvalidParameter, reason);
		}

		// On Windows XP SP3 and newer, using PROV_RSA_AES (instead of PROV_RSA_FULL)
		// would also enable the following hashing algorithms: sha256, sha384, sha512.

		wchar_t const* hashAlgSupp = L"sha1,md5";
		size_t hashAlgSuppLen = wcslen(hashAlgSupp);

		AutoArray<wchar_t> hashAlgBuff(new wchar_t[hashAlgSuppLen + 1]);
		CopyMemory(hashAlgBuff.Get(), hashAlgSupp, sizeof(wchar_t) * (hashAlgSuppLen + 1));

		if (SFS_MatchLists(hashAlgBuff.Get(), hashAlgList))
		{
			wchar_t* delimiter = wcschr(hashAlgBuff.Get(), L',');
			if (delimiter)
				*delimiter = L'\0';

			hashAlg = hashAlgBuff.Get();
			hashAlgBuff.Reset();
		}
		else
		{
			Log log(Log::Failure, 6 + 3);
			log.AddGrp_Param().AddStr(L"path", winPath.c_str());
			if (how) log.AddStr(L"how", how);
			log.AddStr(L"hashAlgProposed", hashAlgList)
			   .AddStr(L"hashAlgSupported", hashAlgSupp)
			   .AddStatusCode(SFS_StatusCode::InvalidParameter);
			log.AddGrp_Error_Exception(L"Unsupported hash algorithms proposed.", false);
			SfsSendLog(log);

			wstring message;
			message += L"Unsupported hash algorithms proposed. Supported hash algorithms: ";
			message += hashAlgSupp;

			throw SfsStatusResponse(SFS_StatusCode::InvalidParameter, message.c_str());
		}
	
		AutoCryptContext cryptProv;
		{
			HCRYPTPROV h;
			if (!CryptAcquireContext(&h, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT))
				throw WinApiError(L"CryptAcquireContext()");

			cryptProv.Set(h);
		}
		
		ALG_ID cryptAlgId = (hashAlg == L"sha1") ? CALG_SHA1 : CALG_MD5;

		AutoCryptHash cryptHash;
		{
			HCRYPTHASH h;
			if (!CryptCreateHash(cryptProv, cryptAlgId, 0, 0, &h))
				throw WinApiError(L"CryptCreateHash()");

			cryptHash.Set(h);
		}
		
		DWORD cryptHashSize;
		DWORD dataLen = sizeof(cryptHashSize);
		if (!CryptGetHashParam(cryptHash, HP_HASHSIZE, (BYTE*) &cryptHashSize, &dataLen, 0))
			throw WinApiError(L"CryptGetHashParam()");

		hashDataCapacity = s_maxReadSize; // max capacity
		hashData.Set(new unsigned char[hashDataCapacity]);

		unsigned __int64 totalToGo = length ? length : -1;
		unsigned __int64 blockToGo = blockSize ? blockSize : -1;
		unsigned __int64 hashBytes = 0;
		unsigned __int64 fileSize = -1; // -1 stands for uknown file size
		{
			ULARGE_INTEGER size;
			if (GetSize(fileHandle, size))
				fileSize = size.QuadPart;
		}
	
		unsigned int chunkSize = s_maxReadSize;
		unsigned __int64 offset = startOffset;
		if (offset > fileSize)
			chunkSize = 1024;
		else
			for	(unsigned __int64 diff = min(totalToGo, fileSize - offset); chunkSize > 1024; chunkSize >>= 1)
				if (diff > (chunkSize >> 1))
					break;

		AutoArray<unsigned char> buffer(new unsigned char[chunkSize]);
		
		bool eofReached = false;
		while (!eofReached && totalToGo)
		{
			unsigned char* data = buffer.Get();
			unsigned int dataSize = (unsigned int) min(chunkSize, totalToGo);
			eofReached = ReadT(fileHandle, offset, data, dataSize);
			
			offset += dataSize;

			do
			{
				unsigned int hashNow = (unsigned int) min<unsigned __int64>(dataSize, min(totalToGo, blockToGo));
				if (hashNow)
				{
					if (!cryptHash.Valid())
					{
						HCRYPTHASH h;
						if (!CryptCreateHash(cryptProv, cryptAlgId, 0, 0, &h))
							throw WinApiError(L"CryptCreateHash()");

						cryptHash.Set(h);
					}

					if (!CryptHashData(cryptHash, data, hashNow, 0))
						throw WinApiError(L"CryptHashData()");
					
					data += hashNow;
					dataSize -= hashNow;
					totalToGo -= hashNow;
					blockToGo -= hashNow;
					hashBytes += hashNow;
				}

				if (!blockToGo || !totalToGo || eofReached)
				{
					if (hashBytes)
					{	
						if (hashDataSize + cryptHashSize > hashDataCapacity)
							throw WStrException(L"Hash data exceeds the maximal allowed size.");

						DWORD dataLen = cryptHashSize;
						if (!CryptGetHashParam(cryptHash, HP_HASHVAL, hashData.Get() + hashDataSize, &dataLen, 0))
							throw WinApiError(L"CryptGetHashParam()");

						hashDataSize += cryptHashSize;
						hashBytes = 0;
						cryptHash.Reset();
					}
					
					if (!blockToGo)
						blockToGo = blockSize;
				}
			}
			while (dataSize && totalToGo);
		}
	}
	catch (WinApiError const& e)
	{		
		unsigned int statusCode = ToSfsStatusCode(e.GetErrorCode());	
		
		Log log(Log::Failure, 7 + 5);
		log.AddGrp_Param().AddStr(L"path", winPath.c_str());
		if (how) log.AddStr(L"how", how);
		log.AddNum(L"offset", startOffset)
		   .AddNum(L"length", length);
		if (!hashAlg.empty())
			log.AddStr(L"hashAlg", hashAlg.c_str());
		log.AddStatusCode(statusCode);
		log.AddGrp_Error_WinApi(e.GetOperation(), false, e.GetErrorCode());
		SfsSendLog(log);

		throw SfsStatusResponse(statusCode);
	}
	catch (WException const& e)
	{
		unsigned int statusCode = SFS_StatusCode::Failure;	
		
		Log log(Log::Failure, 7 + 3);
		log.AddGrp_Param().AddStr(L"path", winPath.c_str());
		if (how) log.AddStr(L"how", how);
		log.AddNum(L"offset", startOffset)
		   .AddNum(L"length", length);
		if (!hashAlg.empty())
			log.AddStr(L"hashAlg", hashAlg.c_str());
		log.AddStatusCode(statusCode);
		log.AddGrp_Error_Exception(e.What(), false);
		SfsSendLog(log);

		throw SfsStatusResponse(statusCode);
	}
	
	Log log(Log::Success, 6);
	log.AddGrp_Param().AddStr(L"path", winPath.c_str());
	if (how) log.AddStr(L"how", how);
	log.AddNum(L"offset", startOffset)
	   .AddNum(L"length", length)
	   .AddStr(L"hashAlg", hashAlg.c_str());
	SfsSendLog(log);

	SFS_CheckFileReply reply;
	reply.m_hashAlgUsed = hashAlg.c_str();
	reply.m_hashDataSize = hashDataSize;
	reply.m_hashDataPtr = hashData.Get();

	SfsSendResponse(SfsResponse(SFS_ResponseType::CheckFileReply, &reply));
}

void SfsWin::HandleSpaceAvail(SFS_Request const* request)
{
	wstring winPath;
	ExtractWinPath(request, winPath);

	// The built-in FlowSfsWin does support this operation.

	SfsSendLog(Log(Log::Failure, 3 + 3)
		.AddGrp_Param().AddStr(L"path", winPath.c_str())
					   .AddStatusCode(SFS_StatusCode::OpUnsupported)
		.AddGrp_Error_Exception(L"The operation is not supported by this provider.", false));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::OpUnsupported, L"The operation is not supported on the provided path."));
}

void SfsWin::HandleFileCopy(SFS_Request const* request)
{	
	// Note that FileCopy will be emulated if not supported by the provider.

	SFS_FileCopy const* copy = (SFS_FileCopy const*) request->m_content;

	wstring winPath[2];
	ExtractWinPaths(request, winPath[0], winPath[1]);
	
	SfsSendLog(Log(Log::Failure, 5 + 3)
		.AddGrp_Param().AddStr(L"sourcePath", winPath[0].c_str())
					   .AddStr(L"destinationPath", winPath[1].c_str())
					   .AddBln(L"overwrite", copy->m_overwrite)
		 			   .AddStatusCode(SFS_StatusCode::OpUnsupported)
		.AddGrp_Error_Exception(L"The operation is not supported.", false));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::OpUnsupported));
}

void SfsWin::HandleDataCopy(SFS_Request const* request)
{
	// Note that DataCopy will be emulated if not supported by the provider.

	SFS_DataCopy const* copy = (SFS_DataCopy const*) request->m_content;

	FsObjMap::iterator it[2] = { m_fsObjs.find(copy->m_readHandle), m_fsObjs.find(copy->m_writeHandle) };
	bool valid[2] = { true, true };
	for (int i = 0; i < 2; ++i)
		if (it[i] == m_fsObjs.end() || !it[i]->second->File())
			valid[i] = false;

	if (!valid[0] || !valid[1])
	{
		wchar_t const* reason;
		if (!valid[0] && !valid[1])	reason = L"The read and write handles are invalid.";
		else if (!valid[0])			reason = L"The read handle is invalid.";
		else						reason = L"The write handle is invalid.";
			
		SfsSendLog(Log(Log::Failure, 2 + 3)
			.AddGrp_Param().AddStatusCode(SFS_StatusCode::InvalidHandle)
			.AddGrp_Error_Exception(reason, false));

		throw SfsStatusResponse(SFS_StatusCode::InvalidHandle);
	}

	SfsSendLog(Log(Log::Failure, 7 + 3)
		.AddGrp_Param().AddStr(L"readPath", it[0]->second->m_winPath.c_str())
					   .AddStr(L"writePath", it[1]->second->m_winPath.c_str())
					   .AddNum(L"readOffset", copy->m_readOffset)
					   .AddNum(L"writeOffset", copy->m_writeOffset)
					   .AddNum(L"length", copy->m_readLength)
					   .AddStatusCode(SFS_StatusCode::OpUnsupported)
		.AddGrp_Error_Exception(L"The operation is not supported.", false));

	SfsSendResponse(SfsStatusResponse(SFS_StatusCode::OpUnsupported));
}

void SfsWin::HandleNativePath(SFS_Request const* request)
{
	SFS_NativePath const* nativePath = (SFS_NativePath const*) request->m_content;
	if (nativePath->m_fromHandle)
	{
		// Return the native path of the handle supplied.

		FsObj* obj = ExtractFsObj(request);

		SfsSendLog(Log(Log::Success, 3)
			.AddGrp_Param().AddStr(L"path", obj->m_winPath.c_str())
						   .AddStr(L"how", L"ByHandle"));

		SFS_Name name;
		ZeroMemory(&name, sizeof(name));
		name.m_fileName = obj->m_winPath.c_str();
		name.m_attrs.m_type = obj->File() ? SFS_Type::Regular : SFS_Type::Directory;

		SfsSendResponse(SfsResponse(SFS_ResponseType::Name, &name));
	}
	else
	{
		// Return the native path of the SFS path supplied.

		wstring winPath;
		ExtractWinPath(request, winPath);

		SfsSendLog(Log(Log::Success, 2)
			.AddGrp_Param().AddStr(L"path", winPath.c_str()));

		SFS_Name name;
		ZeroMemory(&name, sizeof(name));
		name.m_fileName = winPath.c_str();
		name.m_attrs.m_type = SFS_Type::Unknown;

		SfsSendResponse(SfsResponse(SFS_ResponseType::Name, &name));
	}
}


SfsWin::SfsStatusResponse::SfsStatusResponse(unsigned int statusCode, wchar_t const* errorMessage) 
	: SfsResponse(SFS_ResponseType::Status, &m_status)
{
	m_status.m_statusCode = statusCode;
	if (errorMessage)
	{
		m_errorMessageBuffer = errorMessage;
		m_status.m_errorMessage = m_errorMessageBuffer.c_str();
	}
	else
	{
		if (statusCode == SFS_StatusCode::Ok)
			m_status.m_errorMessage = L"";
		else
		{
			m_status.m_errorMessage = SFS_Describe_StatusCode(statusCode);
			if (!m_status.m_errorMessage)
				m_status.m_errorMessage = SFS_Describe_StatusCode(SFS_StatusCode::Failure);
		}
	}

	if (m_status.m_errorMessage[0])
		m_status.m_languageTag = L"en";
	else
		m_status.m_languageTag = L"";
}

SfsWin::SfsHandleResponse::SfsHandleResponse(unsigned __int64 handle, bool createdNewFile)
	: SfsResponse(SFS_ResponseType::Handle, &m_handle)
{
	m_handle.m_handle = handle;
	m_handle.m_createdNewFile = createdNewFile;
}

void SfsWin::SfsSendResponse(SFS_Response const* response)
{
	if (!m_sfsContext->SendResponse(m_sfsContext, response))
		throw SfsContextFailure();
}

void SfsWin::SfsSendLog(SFS_Log const* log)
{
	if (!m_sfsContext->SendLog(m_sfsContext, log))
		throw SfsContextFailure();
}

void SfsWin::SfsSendLogByHandle(unsigned __int64 handle, SFS_Log const* log)
{
	if (!m_sfsContext->SendLogByHandle(m_sfsContext, handle, log))
		throw SfsContextFailure();
}

void SfsWin::SfsAddWaitObject(void* waitObject)
{
	if (!m_sfsContext->AddWaitObject(m_sfsContext, waitObject))
		throw SfsContextFailure();
}

void SfsWin::SfsAddTimer(unsigned int msFromNow)
{
	if (!m_sfsContext->AddTimer(m_sfsContext, msFromNow))
		throw SfsContextFailure();
}

void SfsWin::SfsWait()
{
	if (!m_sfsContext->Wait(m_sfsContext))
		throw SfsContextFailure();
}

void SfsWin::SfsSetException(wchar_t const* what)
{
	if (!m_sfsContext->SetException(m_sfsContext, what))
		throw SfsContextFailure();
}

void SfsWin::SfsAcquireAsyncContext(SFS_Context** asyncContext)
{
	if (!m_sfsContext->AcquireAsyncContext(m_sfsContext, asyncContext))
		throw SfsContextFailure();
}

void SfsWin::SfsReleaseAsyncContext()
{
	if (!m_sfsContext->ReleaseAsyncContext(m_sfsContext))
		throw SfsContextFailure();
}


SfsWin::FsFile::FsFile() 
	: m_fileHandle(0)
	, m_fileAccess(0)
	, m_append(false)
	, m_textMode(false)
	, m_textOffset(0)
{}


void SfsWin::AddFsObj(FsObj* fsObj)
{
	fsObj->m_sfsHandle = ++m_lastSfsHandle;
	if (!m_fsObjs.insert(FsObjMap::value_type(fsObj->m_sfsHandle, fsObj)).second)
		throw WStrException(L"SfsWin: Adding FS object failed: The associated handle is already used.");
}

void SfsWin::DelFsObj(unsigned __int64 sfsHandle)
{
	FsObjMap::iterator it = m_fsObjs.find(sfsHandle);
	if (it == m_fsObjs.end())
		throw WStrException(L"SfsWin: Removing FS object failed: The provided handle does not exist.");

	delete it->second;
	m_fsObjs.erase(it);
}


bool SfsWin::GetSize(HANDLE handle, ULARGE_INTEGER& size)
{
	size.LowPart = GetFileSize(handle, &size.HighPart);
	if (size.LowPart == INVALID_FILE_SIZE)
		if (GetLastError() != NO_ERROR)
			return false;

	return true;
}


void SfsWin::GetFileInfoT(HANDLE handle, BY_HANDLE_FILE_INFORMATION& fileInfo)
{
	if (!GetFileInformationByHandle(handle, &fileInfo))
		throw WinApiError(L"GetFileInformationByHandle()");
}

void SfsWin::GetFileInfoT(wchar_t const* winPath, BY_HANDLE_FILE_INFORMATION& fileInfo)
{
	DWORD winAttrs = GetFileAttributes(winPath);
	if (winAttrs == INVALID_FILE_ATTRIBUTES)
		throw WinApiError(L"GetFileAttributes()");

	DWORD flags = 0;
	if (winAttrs & FILE_ATTRIBUTE_DIRECTORY)
		flags |= FILE_FLAG_BACKUP_SEMANTICS;
						
	AutoHandle handle(OpenExistingT(winPath, FILE_READ_ATTRIBUTES, flags));
	GetFileInfoT(handle, fileInfo);
}

void SfsWin::GetFileInfoT(wchar_t const* winPath, WIN32_FILE_ATTRIBUTE_DATA& fileInfo)
{
	if (!GetFileAttributesEx(winPath, GetFileExInfoStandard, &fileInfo))
		throw WinApiError(L"GetFileAttributesEx()");
}



bool SfsWin::IsFile(WIN32_FIND_DATA const& findData)
{
	return !(findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
}

bool SfsWin::IsFile(BY_HANDLE_FILE_INFORMATION const& fileInfo)
{
	return !(fileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
}

bool SfsWin::IsFile(wchar_t const* winPath)
{
	DWORD winAttrs = GetFileAttributes(winPath);
	if (winAttrs != INVALID_FILE_ATTRIBUTES)
		if (!(winAttrs & FILE_ATTRIBUTE_DIRECTORY))
			return true;

	return false;
}

bool SfsWin::IsDirectory(wchar_t const* winPath)
{
	DWORD winAttrs = GetFileAttributes(winPath);
	if (winAttrs != INVALID_FILE_ATTRIBUTES)
		if (winAttrs & FILE_ATTRIBUTE_DIRECTORY)
			return true;

	return false;
}


bool SfsWin::FindFileFilter(WIN32_FIND_DATA const& findData)
{
	if (!wcscmp(findData.cFileName, L".") ||
		!wcscmp(findData.cFileName, L".."))
		return false;

	return true;
}


HANDLE SfsWin::OpenExistingT(wchar_t const* winPath, DWORD access, DWORD flags)
{
	HANDLE handle = CreateFile(winPath, access, 
		FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
		0, OPEN_EXISTING, flags, 0);

	if (handle == INVALID_HANDLE_VALUE)
		throw WinApiError(L"CreateFile()");

	return handle;
}


void SfsWin::SetSizeT(HANDLE handle, unsigned __int64 size)
{
	LARGE_INTEGER pos;
	pos.QuadPart = size;
	pos.LowPart = SetFilePointer(handle, pos.LowPart, &pos.HighPart, FILE_BEGIN);
	if (pos.LowPart == INVALID_SET_FILE_POINTER && GetLastError() != NO_ERROR)
		throw WinApiError(L"SetFilePointer()");

	if (!SetEndOfFile(handle))
		throw WinApiError(L"SetEndOfFile()");
}

void SfsWin::SetTimesT(HANDLE handle, WinTimes const& times)
{
	if (!SetFileTime(handle, times.m_createTime.Get(), times.m_accessTime.Get(), times.m_modifyTime.Get()))
		throw WinApiError(L"SetFileTime");
}


bool SfsWin::ReadT(HANDLE handle, unsigned __int64 offset, byte* data, unsigned int& size)
{
	bool eofReached = false;
	unsigned int totalRead = 0;
	
	while (!eofReached && totalRead != size)
	{
		ULARGE_INTEGER pos;
		pos.QuadPart = offset + totalRead;
		m_overlapped.Offset = pos.LowPart;
		m_overlapped.OffsetHigh = pos.HighPart;
		ResetEvent(m_overlapped.hEvent);

		DWORD read;
		if (ReadFile(handle, data + totalRead, size - totalRead, &read, &m_overlapped))
			totalRead += read;
		else if (GetLastError() == ERROR_HANDLE_EOF)
			eofReached = true;
		else if (GetLastError() != ERROR_IO_PENDING)
			throw WinApiError(L"ReadFile()");
		else
		{
			while (true)
			{
				try
				{
					SfsAddWaitObject(m_overlapped.hEvent);
					SfsWait();
				}
				catch (...)
				{
					CancelIo(handle);
					WaitForSingleObject(m_overlapped.hEvent, INFINITE);
					throw;
				}

				if (GetOverlappedResult(handle, &m_overlapped, &read, false))
				{
					totalRead += read;
					break;
				}
					
				if (GetLastError() == ERROR_HANDLE_EOF)
				{
					eofReached = true;
					break;	
				}

				if (GetLastError() != ERROR_IO_INCOMPLETE)
					throw WinApiError(L"GetOverlappedResult()");
			}
		}
	}

	size = totalRead;
	return eofReached;
}

void SfsWin::WriteT(HANDLE handle, unsigned __int64 offset, byte const* data, unsigned int size)
{
	while (size)
	{
		ULARGE_INTEGER pos;
		pos.QuadPart = offset;
		m_overlapped.Offset = pos.LowPart;
		m_overlapped.OffsetHigh = pos.HighPart;
		ResetEvent(m_overlapped.hEvent);

		DWORD written;
		if (!WriteFile(handle, data, size, &written, &m_overlapped))
		{
			if (GetLastError() != ERROR_IO_PENDING)
				throw WinApiError(L"WriteFile()");

			while (true)
			{
				try
				{
					SfsAddWaitObject(m_overlapped.hEvent);
					SfsWait();
				}
				catch (...)
				{
					CancelIo(handle);
					WaitForSingleObject(m_overlapped.hEvent, INFINITE);
					throw;
				}

				if (GetOverlappedResult(handle, &m_overlapped, &written, false))
					break;
				
				if (GetLastError() != ERROR_IO_INCOMPLETE)
					throw WinApiError(L"GetOverlappedResult()");
			}
		}

		data += written;
		size -= written;
		if (offset != -1)
			offset += written;
	}
}


void SfsWin::SfsToWinPath(wchar_t const* sfsPath, wstring& winPath) const
{  
	size_t sfsPathLen = sfsPath ? wcslen(sfsPath) : 0;
	if (sfsPathLen == 0 || sfsPath[0] != L'/')
		throw WStrException(L"SfsWin: Unexpected SFS path received: Expecting absolute path.");

	if (sfsPathLen == 1)
		winPath.resize(0);
	else 
	{
		if (sfsPathLen == 2)
		{
			// '/?' paths are converted to '?:\'
			
			winPath.resize(3);
			winPath[0] = sfsPath[1];
			winPath[1] = L':';
			winPath[2] = L'\\';
		}
		else // if (sfsPathLen >= 3)
		{
			// /?/* paths are converted to local ?:\*
			// /??* paths are converted to network \\??*
			
			if (sfsPath[2] != L'/')
			{
				winPath.reserve(sfsPathLen + 1);
				winPath = L"\\\\";
				winPath += sfsPath + 1;
			}
			else
			{
				winPath.reserve(sfsPathLen);
				winPath.resize(2);
				winPath[0] = sfsPath[1];
				winPath[1] = L':';
				winPath += sfsPath + 2;
			}
					
			// Replace '/' delimiters with '\'
			for (size_t i = 0; i < winPath.size(); ++i)
				if (winPath[i] == L'/')
					winPath[i] = L'\\';
		}
	}
}


wchar_t const* SfsWin::ExtractSfsPath(SFS_Request const* request)
{
	switch (request->m_type)
	{
	default:
		throw WStrException(L"SfsWin: Requested path extraction for " + ToStringRequestType(request->m_type) + L" request.");

	case SFS_RequestType::Open:		return ((SFS_Open const*) request->m_content)->m_fileName;
	case SFS_RequestType::SetStat:	return ((SFS_SetStat const*) request->m_content)->m_path;
	case SFS_RequestType::OpenDir:	return ((SFS_OpenDir const*) request->m_content)->m_path;
	case SFS_RequestType::Remove:	return ((SFS_Remove const*) request->m_content)->m_fileName;
	case SFS_RequestType::MkDir:	return ((SFS_MkDir const*) request->m_content)->m_path;
	case SFS_RequestType::RmDir:	return ((SFS_RmDir const*) request->m_content)->m_path;
	case SFS_RequestType::RealPath:	return ((SFS_RealPath const*) request->m_content)->m_path;
	case SFS_RequestType::Stat:		return ((SFS_Stat const*) request->m_content)->m_path;
	case SFS_RequestType::ReadLink:	return ((SFS_ReadLink const*) request->m_content)->m_path;
	case SFS_RequestType::CheckFileName:	return ((SFS_CheckFileName const*) request->m_content)->m_fileName;
	case SFS_RequestType::SpaceAvail: return ((SFS_SpaceAvail const*) request->m_content)->m_path;
	case SFS_RequestType::NativePath: return ((SFS_NativePath const*) request->m_content)->m_path;
	}
}


void SfsWin::ExtractWinPath(SFS_Request const* request, wstring& winPath)
{
	SfsToWinPath(ExtractSfsPath(request), winPath);
}

void SfsWin::ExtractWinPaths(SFS_Request const* request, wstring& winPath1, wstring& winPath2)
{
	switch (request->m_type)
	{
	default:
		throw WStrException(L"SfsWin: Requested paths extraction for " + ToStringRequestType(request->m_type)  + L" request.");

	case SFS_RequestType::Rename:
		{
			SFS_Rename const* rename = (SFS_Rename const*) request->m_content;
			SfsToWinPath(rename->m_oldPath, winPath1);
			SfsToWinPath(rename->m_newPath, winPath2);
		}
		break;

	case SFS_RequestType::Link:
		{
			SFS_Link const* link = (SFS_Link const*) request->m_content;
			SfsToWinPath(link->m_newLinkPath, winPath1);
			SfsToWinPath(link->m_existingPath, winPath2);
		}
		break;

	case SFS_RequestType::FileCopy:
		{
			SFS_FileCopy const* copy = (SFS_FileCopy const*) request->m_content;
			SfsToWinPath(copy->m_srcFileName, winPath1);
			SfsToWinPath(copy->m_dstFileName, winPath2);
		}
		break;
	}
}


unsigned __int64 SfsWin::ExtractSfsHandle(SFS_Request const* request)
{
	switch (request->m_type)
	{
	default:
		throw WStrException(L"SfsWin: Requested handle extraction for " + ToStringRequestType(request->m_type) + L" request.");
		
	case SFS_RequestType::Close:	return ((SFS_Close const*) request->m_content)->m_handle;
	case SFS_RequestType::Read:		return ((SFS_Read const*) request->m_content)->m_handle;
	case SFS_RequestType::Write:	return ((SFS_Write const*) request->m_content)->m_handle;
	case SFS_RequestType::FStat:	return ((SFS_FStat const*) request->m_content)->m_handle;
	case SFS_RequestType::FSetStat:	return ((SFS_FSetStat const*) request->m_content)->m_handle;
	case SFS_RequestType::ReadDir:	return ((SFS_ReadDir const*) request->m_content)->m_handle;
	case SFS_RequestType::Block:	return ((SFS_Block const*) request->m_content)->m_handle;
	case SFS_RequestType::Unblock:	return ((SFS_Unblock const*) request->m_content)->m_handle;
	case SFS_RequestType::TextSeek:	return ((SFS_TextSeek const*) request->m_content)->m_handle;
	case SFS_RequestType::CheckFileHandle:	return ((SFS_CheckFileHandle const*) request->m_content)->m_handle;
	case SFS_RequestType::NativePath:	return ((SFS_NativePath const*) request->m_content)->m_handle;
	}	
}


SfsWin::FsObj* SfsWin::ExtractFsObj(SFS_Request const* request)
{
	FsObjMap::iterator it = m_fsObjs.find(ExtractSfsHandle(request));
	if (it == m_fsObjs.end())
	{
		unsigned int statusCode = SFS_StatusCode::InvalidHandle;
		
		Log log(Log::Failure, 3 + 3);
		log.AddGrp_Param();
		
		switch (request->m_type)
		{
		case SFS_RequestType::FStat:
		case SFS_RequestType::FSetStat:
		case SFS_RequestType::CheckFileHandle:
		case SFS_RequestType::NativePath:
			log.AddStr(L"how", L"ByHandle");
		}

		log.AddStatusCode(statusCode);
		log.AddGrp_Error_StatusCode(statusCode);
		SfsSendLog(log);

		throw SfsStatusResponse(statusCode);
	}

	return it->second;
}

SfsWin::FsFile* SfsWin::ExtractFsFile(SFS_Request const* request)
{
	SfsWin::FsObj* obj = ExtractFsObj(request);
	if (!obj->File())
	{
		unsigned int statusCode = SFS_StatusCode::InvalidHandle;

		SfsSendLog(Log(Log::Failure, 2 + 3)
			.AddGrp_Param().AddStatusCode(statusCode)
			.AddGrp_Error_StatusCode(statusCode));

		throw SfsStatusResponse(statusCode);
	}

	return obj->File();
};

SfsWin::FsObj* SfsWin::ExtractFsRootOrDir(SFS_Request const* request)
{
	SfsWin::FsObj* obj = ExtractFsObj(request);
	if (obj->File())
	{
		unsigned int statusCode = SFS_StatusCode::InvalidHandle;

		SfsSendLog(Log(Log::Failure, 2 + 3)
			.AddGrp_Param().AddStatusCode(statusCode)
			.AddGrp_Error_StatusCode(statusCode));

		throw SfsStatusResponse(statusCode);
	}

	return obj;
};


SFS_Attrs const* SfsWin::ExtractSfsAttrs(SFS_Request const* request)
{
	switch (request->m_type)
	{
	default:
		throw WStrException(L"SfsWin::ExtractSfsAttrs: Requested attributes extraction for " + ToStringRequestType(request->m_type) + L" request.");

	case SFS_RequestType::Open:
		return &((SFS_Open const*) request->m_content)->m_attrs;

	case SFS_RequestType::SetStat:
		return &((SFS_SetStat const*) request->m_content)->m_attrs;

	case SFS_RequestType::FSetStat:
		return &((SFS_FSetStat const*) request->m_content)->m_attrs;

	case SFS_RequestType::MkDir:
		return &((SFS_MkDir const*) request->m_content)->m_attrs;
	}
}

void SfsWin::ExtractWinAttrs_PrepareLog(SFS_Request const* request, SFS::Log& log)
{
	log.AddGrp_Param();
	if (request->m_type == SFS_RequestType::FSetStat)
	{
		log.AddStr(L"path", ExtractFsObj(request)->m_winPath.c_str());
		log.AddStr(L"how", L"ByHandle");
	}
	else
	{
		wstring winPath;
		ExtractWinPath(request, winPath);
		log.AddStr(L"path", winPath.c_str(), Log::CopyValue);
	}
}

wchar_t const* SfsWin::ExtractWinAttrs_GetWhileSuffix(SFS_Request const* request)
{
	switch (request->m_type)
	{
	default:						return L"";	
	case SFS_RequestType::Open:		return L"while opening file.";
	case SFS_RequestType::SetStat:	
	case SFS_RequestType::FSetStat:	return L"while settings file status.";
	case SFS_RequestType::MkDir:	return L"while creating directory.";
	}
}


void SfsWin::ExtractWinTimes(SFS_Request const* request, SfsWin::WinTimes& winTimes)
{
	SFS_Attrs const* sfsAttrs = ExtractSfsAttrs(request);

	winTimes.m_createTime.m_present = false;
	winTimes.m_accessTime.m_present = false;
	winTimes.m_modifyTime.m_present = false;

	if (sfsAttrs->m_validAttrFlags & (SFS_Attr::CreateTime | SFS_Attr::ModifyTime | SFS_Attr::AccessTime))
	{
		static wchar_t const* const convertTimeFailurePrefix = L"Converting time to the native form failed ";
		static wchar_t const* const convertTimeErrorMessage = L"The provided time is invalid or it cannot be represented in the native form.";
	
		bool ns = !!(sfsAttrs->m_validAttrFlags & SFS_Attr::Subseconds);

		if (sfsAttrs->m_validAttrFlags & SFS_Attr::CreateTime)
		{
			if (!ToWinTime(sfsAttrs->m_createTime, ns ? sfsAttrs->m_createTimeNs : 0, winTimes.m_createTime.m_time))
			{
				wstring message = wstring(convertTimeFailurePrefix) + ExtractWinAttrs_GetWhileSuffix(request);

				Log log(SFS_LogAction::Default + ConvertTimeFailure, message.c_str(), false, 7 + 3);
				ExtractWinAttrs_PrepareLog(request, log);
				log.AddStr(L"timeOrigin", L"CreateTime");
				log.AddNum(L"timeSeconds", sfsAttrs->m_createTime);
				log.AddNum(L"timeNanosecods", ns ? sfsAttrs->m_createTimeNs : 0);
				log.AddStatusCode(SFS_StatusCode::InvalidParameter);
				log.AddGrp_Error_Exception(convertTimeErrorMessage, false);
				SfsSendLog(log);

				throw SfsStatusResponse(SFS_StatusCode::InvalidParameter, L"Failed to convert CreateTime to the native form.");
			}

			winTimes.m_createTime.m_present = true;
		}
		
		if (sfsAttrs->m_validAttrFlags & SFS_Attr::ModifyTime)
		{
			if (!ToWinTime(sfsAttrs->m_modifyTime, ns ? sfsAttrs->m_modifyTimeNs : 0, winTimes.m_modifyTime.m_time))
			{
				wstring message = wstring(convertTimeFailurePrefix) + ExtractWinAttrs_GetWhileSuffix(request);

				Log log(SFS_LogAction::Default + ConvertTimeFailure, message.c_str(), false, 7 + 3);
				ExtractWinAttrs_PrepareLog(request, log);
				log.AddStr(L"timeOrigin", L"ModifyTime");
				log.AddNum(L"timeSeconds", sfsAttrs->m_modifyTime);
				log.AddNum(L"timeNanosecods", ns ? sfsAttrs->m_modifyTimeNs : 0);
				log.AddStatusCode(SFS_StatusCode::InvalidParameter);
				log.AddGrp_Error_Exception(convertTimeErrorMessage, false);
				SfsSendLog(log);

				throw SfsStatusResponse(SFS_StatusCode::InvalidParameter, L"Failed to convert ModifyTime to the native form.");
			}

			winTimes.m_modifyTime.m_present = true;
		}

		if (sfsAttrs->m_validAttrFlags & SFS_Attr::AccessTime)
		{
			if (!ToWinTime(sfsAttrs->m_accessTime, ns ? sfsAttrs->m_accessTimeNs : 0, winTimes.m_accessTime.m_time))
			{
				wstring message = wstring(convertTimeFailurePrefix)  + ExtractWinAttrs_GetWhileSuffix(request);

				Log log(SFS_LogAction::Default + ConvertTimeFailure, message.c_str(), false, 7 + 3);
				ExtractWinAttrs_PrepareLog(request, log);
				log.AddStr(L"timeOrigin", L"AccessTime");
				log.AddNum(L"timeSeconds", sfsAttrs->m_accessTime);
				log.AddNum(L"timeNanosecods", ns ? sfsAttrs->m_accessTimeNs : 0);
				log.AddStatusCode(SFS_StatusCode::InvalidParameter);
				log.AddGrp_Error_Exception(convertTimeErrorMessage, false);
				SfsSendLog(log);

				throw SfsStatusResponse(SFS_StatusCode::InvalidParameter, L"Failed to convert AccessTime to the native form.");
			}

			winTimes.m_accessTime.m_present = true;
		}
	}
}


void SfsWin::ToSfsAttrs(WIN32_FIND_DATA const& data, SFS_Attrs& sfsAttrs)
{
	WIN32_FILE_ATTRIBUTE_DATA attrData;
	attrData.dwFileAttributes = data.dwFileAttributes;
	attrData.ftCreationTime = data.ftCreationTime;
	attrData.ftLastAccessTime = data.ftLastAccessTime;
	attrData.ftLastWriteTime = data.ftLastWriteTime;
	attrData.nFileSizeHigh = data.nFileSizeHigh;
	attrData.nFileSizeLow = data.nFileSizeLow;
	ToSfsAttrs(attrData, sfsAttrs);
}

void SfsWin::ToSfsAttrs(WIN32_FILE_ATTRIBUTE_DATA const& data, SFS_Attrs& sfsAttrs)
{	
	sfsAttrs.m_validAttrFlags = SFS_Attr::Size | SFS_Attr::AccessTime | SFS_Attr::CreateTime |
								SFS_Attr::ModifyTime | SFS_Attr::Subseconds | SFS_Attr::Bits;
	
	if (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		sfsAttrs.m_type = SFS_Type::Directory;
	else
		sfsAttrs.m_type = SFS_Type::Regular;
	
	sfsAttrs.m_size = data.nFileSizeHigh;
	sfsAttrs.m_size <<= 32;
	sfsAttrs.m_size |= data.nFileSizeLow;
	
	ToSfsTime(data.ftCreationTime, sfsAttrs.m_createTime, sfsAttrs.m_createTimeNs);
	ToSfsTime(data.ftLastAccessTime, sfsAttrs.m_accessTime, sfsAttrs.m_accessTimeNs);
	ToSfsTime(data.ftLastWriteTime, sfsAttrs.m_modifyTime, sfsAttrs.m_modifyTimeNs);

	sfsAttrs.m_attrBitsValid = SFS_AttrBit::CaseInsensitive | SFS_AttrBit::ReadOnly | SFS_AttrBit::System |
							   SFS_AttrBit::Hidden | SFS_AttrBit::Archive | SFS_AttrBit::Encrypted |
							   SFS_AttrBit::Compressed | SFS_AttrBit::Sparse;

	sfsAttrs.m_attrBits = SFS_AttrBit::CaseInsensitive;
	if (data.dwFileAttributes & FILE_ATTRIBUTE_READONLY)	sfsAttrs.m_attrBits |= SFS_AttrBit::ReadOnly;
	if (data.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM)		sfsAttrs.m_attrBits |= SFS_AttrBit::System;
	if (data.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN)		sfsAttrs.m_attrBits |= SFS_AttrBit::Hidden;
	if (data.dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE)		sfsAttrs.m_attrBits |= SFS_AttrBit::Archive;
	if (data.dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED)	sfsAttrs.m_attrBits |= SFS_AttrBit::Encrypted;
	if (data.dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED)	sfsAttrs.m_attrBits |= SFS_AttrBit::Compressed;
	if (data.dwFileAttributes & FILE_ATTRIBUTE_SPARSE_FILE)	sfsAttrs.m_attrBits |= SFS_AttrBit::Sparse;
}

void SfsWin::ToSfsAttrs(BY_HANDLE_FILE_INFORMATION const& info, SFS_Attrs& sfsAttrs)
{	
	WIN32_FILE_ATTRIBUTE_DATA data;
	data.dwFileAttributes = info.dwFileAttributes;
	data.ftCreationTime = info.ftCreationTime;
	data.ftLastAccessTime = info.ftLastAccessTime;
	data.ftLastWriteTime = info.ftLastWriteTime;
	data.nFileSizeHigh = info.nFileSizeHigh;
	data.nFileSizeLow = info.nFileSizeLow;
	ToSfsAttrs(data, sfsAttrs);

	sfsAttrs.m_validAttrFlags |= SFS_Attr::LinkCount;
	sfsAttrs.m_linkCount = info.nNumberOfLinks;
}


DWORD SfsWin::RemoveReadOnlyWinAttrBits(DWORD winAttrs)
{
	winAttrs &= ~((DWORD) (
		FILE_ATTRIBUTE_ARCHIVE |
		FILE_ATTRIBUTE_HIDDEN |
		FILE_ATTRIBUTE_NORMAL | 
		FILE_ATTRIBUTE_NOT_CONTENT_INDEXED |
		FILE_ATTRIBUTE_OFFLINE |
		FILE_ATTRIBUTE_READONLY |
		FILE_ATTRIBUTE_SYSTEM |
		FILE_ATTRIBUTE_TEMPORARY));

	return winAttrs;
}

DWORD SfsWin::ModifyBasicWinAttrBits(DWORD winAttrs, SFS_Attrs const* sfsAttrs)
{
	if (sfsAttrs->m_validAttrFlags & SFS_Attr::Bits)
	{
		if (sfsAttrs->m_attrBitsValid & SFS_AttrBit::ReadOnly)
			if (sfsAttrs->m_attrBits & SFS_AttrBit::ReadOnly)
				winAttrs |= FILE_ATTRIBUTE_READONLY;
			else
				winAttrs &= ~((DWORD) FILE_ATTRIBUTE_READONLY);
		
		if (sfsAttrs->m_attrBitsValid & SFS_AttrBit::System)
			if (sfsAttrs->m_attrBits & SFS_AttrBit::System)
				winAttrs |= FILE_ATTRIBUTE_SYSTEM;
			else
				winAttrs &= ~((DWORD) FILE_ATTRIBUTE_SYSTEM);
			
		if (sfsAttrs->m_attrBitsValid & SFS_AttrBit::Hidden)
			if (sfsAttrs->m_attrBits & SFS_AttrBit::Hidden)
				winAttrs |= FILE_ATTRIBUTE_HIDDEN;
			else
				winAttrs &= ~((DWORD) FILE_ATTRIBUTE_HIDDEN);

		if (sfsAttrs->m_attrBitsValid & SFS_AttrBit::Archive)
			if (sfsAttrs->m_attrBits & SFS_AttrBit::Archive)
				winAttrs |= FILE_ATTRIBUTE_ARCHIVE;
			else
				winAttrs &= ~((DWORD) FILE_ATTRIBUTE_ARCHIVE);
	}

	return winAttrs;
}


bool SfsWin::ToWinTime(unsigned __int64 sfsTime, unsigned int sfsTimeNs, FILETIME& winTime)
{
	if (sfsTimeNs < 1000000000 && UxtimeToFtime(sfsTime, winTime))
	{
		winTime.dwLowDateTime += sfsTimeNs / 100;
		return true;
	}
	
	return false;
}

bool SfsWin::ToSfsTime(FILETIME const& winTime, unsigned __int64& sfsTime, unsigned int& sfsTimeNs)
{
	time_t uxTime;
	if (FtimeToUxtime(winTime, uxTime))
	{
		sfsTime = uxTime;
		sfsTimeNs = (winTime.dwLowDateTime % 10000000) * 100;
		return true;	
	}
	
	return false;
}


DWORD SfsWin::ToWinAccessRights(unsigned int aceMask)
{
	DWORD ret = 0;

	if (aceMask & SFS_AceMask::Read)				ret |= FILE_READ_DATA;
	if (aceMask & SFS_AceMask::Write)				ret |= FILE_WRITE_DATA;
	if (aceMask & SFS_AceMask::Append)				ret |= FILE_APPEND_DATA;
	if (aceMask & SFS_AceMask::ReadNamedAttrs)		ret |= FILE_READ_EA;
	if (aceMask & SFS_AceMask::WriteNamedAttrs)		ret |= FILE_WRITE_EA;
	if (aceMask & SFS_AceMask::Execute)				ret |= FILE_EXECUTE;
	if (aceMask & SFS_AceMask::DeleteChild)			ret |= FILE_DELETE_CHILD;
	if (aceMask & SFS_AceMask::ReadAttrs)			ret |= FILE_READ_ATTRIBUTES;
	if (aceMask & SFS_AceMask::WriteAttrs)			ret |= FILE_WRITE_ATTRIBUTES;
	if (aceMask & SFS_AceMask::Delete)				ret |= DELETE;
	if (aceMask & SFS_AceMask::ReadAcl)				ret |= READ_CONTROL;
	if (aceMask & SFS_AceMask::WriteAcl)			ret |= WRITE_DAC;
	if (aceMask & SFS_AceMask::WriteOwner)			ret |= WRITE_OWNER;
	if (aceMask & SFS_AceMask::Synchronize)			ret |= SYNCHRONIZE;
	if (aceMask & SFS_AceMask::AccessAuditAlarm)	ret |= ACCESS_SYSTEM_SECURITY;

	return ret;
}


unsigned int SfsWin::ToSfsStatusCode(DWORD winError)
{
	switch (winError)
	{
	default:						return SFS_StatusCode::Failure;
	
	case ERROR_NO_MORE_FILES:		return SFS_StatusCode::Eof;
	case ERROR_HANDLE_EOF:			return SFS_StatusCode::Eof;
	case ERROR_FILE_NOT_FOUND:		return SFS_StatusCode::NoSuchFile;
	case ERROR_ACCESS_DENIED:		return SFS_StatusCode::PermissionDenied;

	case ERROR_NOT_SUPPORTED:		return SFS_StatusCode::OpUnsupported;
	
	// Windows handles are already verified before issuing any WinAPI calls:
	// so ERROR_INVALID_HANDLE would indicate some kind of internal problem
	// for which the generic SFS_StatusCode::Failure reply is more suitable.
	// case ERROR_INVALID_HANDLE:	return SFS_StatusCode::InvalidHandle;
	case ERROR_PATH_NOT_FOUND:		return SFS_StatusCode::NoSuchPath;
	case ERROR_INVALID_DRIVE:		return SFS_StatusCode::NoSuchPath;
	case ERROR_FILE_EXISTS:			return SFS_StatusCode::FileAlreadyExists;
	case ERROR_ALREADY_EXISTS:		return SFS_StatusCode::FileAlreadyExists;
	case ERROR_WRITE_PROTECT:		return SFS_StatusCode::WriteProtect;
	case ERROR_NO_MEDIA_IN_DRIVE:	return SFS_StatusCode::NoMedia;
	case ERROR_MEDIA_NOT_AVAILABLE:	return SFS_StatusCode::NoMedia;
	case ERROR_NOT_READY:			return SFS_StatusCode::NoMedia;
	case ERROR_DISK_FULL:			return SFS_StatusCode::NoSpaceOnFilesystem;
	case ERROR_HANDLE_DISK_FULL:	return SFS_StatusCode::NoSpaceOnFilesystem;
	
	case ERROR_SHARING_VIOLATION:	return SFS_StatusCode::LockConflict;
	case ERROR_LOCK_VIOLATION:		return SFS_StatusCode::LockConflict;
	case ERROR_DIR_NOT_EMPTY:		return SFS_StatusCode::DirNotEmpty;
	case ERROR_DIRECTORY:			return SFS_StatusCode::NotADirectory;

	case ERROR_DELETE_PENDING:		return SFS_StatusCode::DeletePending;
	case ERROR_FILE_CORRUPT:		return SFS_StatusCode::FileCorrupt;
	case ERROR_DISK_CORRUPT:		return SFS_StatusCode::FileCorrupt;
	}
}

wstring SfsWin::ToStringRequestType(unsigned int requestType)
{
	wchar_t const* s = SFS_ToString_RequestType(requestType);
	if (s)
		return s;
	else
	{
		wstringstream ss;
		ss << requestType;
		return ss.str();
	}
}

	