#pragma once

// WinAPI
#include <windows.h>

// STL
#include <map>
#include <queue>
#include <memory>
#include <string>

// SfsProvider
#include "SfsProvider.h"
#include "SfsUtilities.h"

// SfsWin
#include "SfsWinUtilities.h"


// SfsWin is a sample SFS provider based on the built-in FlowSfsWin provider.
// Compared to the FlowSfsWin, this sample lacks the folowing features:
// - Support for RootPath or any other custom provider settings;
// - Support for file security attributes: owner, group, POSIX permissions, ACL;
// - Support for symbolic and hard links;
// - Ability to modify encryption, compression, and sparse file attribute bits;
// - Ability to generate text hint in the SFS_Stat and SFS_FStat requests;
// - Ability to create a multi-level directory tree with one SFS_MkDir request; 
// - Support for the SFS_SpaceAvail request;
// - Support for sha256, sha384, and sha512 hashing algorithms;
// - Support for Windows paths longer than 255 characters.

class SfsWin
{
public:
	SfsWin() : m_lastSfsHandle(0) { m_overlapped.hEvent = 0; }
	~SfsWin();

	bool HandleInit(SFS_Init const* init, SFS_Context* context);
	void HandleRequest(SFS_Request const* request, SFS_Context* context);
	void HandleDestroy(SFS_Context* context);

	void HandleOpen(SFS_Request const* request);
	void HandleClose(SFS_Request const* request);
	void HandleRead(SFS_Request const* request);
	void HandleWrite(SFS_Request const* request);
   	void HandleStat(SFS_Request const* request);
	void HandleSetStat(SFS_Request const* request);
	void HandleOpenDir(SFS_Request const* request);
	void HandleReadDir(SFS_Request const* request);
	void HandleRemove(SFS_Request const* request);
	void HandleMkDir(SFS_Request const* request);
	void HandleRmDir(SFS_Request const* request);
	void HandleRealPath(SFS_Request const* request);
	void HandleRename(SFS_Request const* request);
	void HandleReadLink(SFS_Request const* request);
	void HandleLink(SFS_Request const* request);
	void HandleBlock(SFS_Request const* request);
	void HandleUnblock(SFS_Request const* request);
	void HandleTextSeek(SFS_Request const* request);
    void HandleCheckFile(SFS_Request const* request);
	void HandleSpaceAvail(SFS_Request const* request);
	void HandleFileCopy(SFS_Request const* request);
	void HandleDataCopy(SFS_Request const* request);
	void HandleNativePath(SFS_Request const* request);
	
private:
	class SfsResponse : public SFS_Response
	{
	public:
		SfsResponse(unsigned int type, void const* content) { m_type = type; m_content = content; }
		operator SFS_Response const*() const { return (SFS_Response const*) this; }
	};

	class SfsStatusResponse : public SfsResponse
	{
	public:
		SfsStatusResponse(unsigned int statusCode, wchar_t const* errorMessage = 0);
	private:
		SFS_Status m_status;
		std::wstring m_errorMessageBuffer;
	};

	class SfsHandleResponse : public SfsResponse
	{
	public:
		SfsHandleResponse(unsigned __int64 handle, bool createdNewFile = false);
	private:
		SFS_Handle m_handle;
	};

	class SfsContextFailure {};

	// The following functions can throw SfsContextFailure.
	void SfsSendResponse(SFS_Response const* response);
	void SfsSendLog(SFS_Log const* log);
	void SfsSendLogByHandle(unsigned __int64 handle, SFS_Log const* log);
	void SfsAddWaitObject(void* waitObject);
	void SfsAddTimer(unsigned int msFromNow);
	void SfsWait();
	void SfsSetException(wchar_t const* what);
	void SfsAcquireAsyncContext(SFS_Context** asyncContext);
	void SfsReleaseAsyncContext();

	SFS_Context* m_sfsContext;

private:
	class FsRoot;
	class FsDir;
	class FsFile;

	class FsObj : public NoCopy
	{
	protected:
		FsObj() {}

	public:
		virtual ~FsObj() {}

		virtual FsRoot* Root() { return 0; }
		virtual FsDir*  Dir() { return 0; }
		virtual FsFile* File() { return 0; }
			
		static bool ValidHandle(HANDLE h) { return h && h != INVALID_HANDLE_VALUE; }

		std::wstring m_sfsPath;
		std::wstring m_winPath;
		unsigned __int64 m_sfsHandle;
	};

	class FsRoot : public FsObj
	{
	public:
		FsRoot() : m_drives(0), m_nextDrive(L'A') {}

		FsRoot* Root() { return this; }

		DWORD m_drives; // drive mask, see GetLogicalDrives()
		wchar_t m_nextDrive; // next drive letter candidate
		
		SFS::ListDir m_listDir;
	};

	class FsDir : public FsObj
	{
	public:
		FsDir() : m_findHandle(0), m_noMoreFiles(false), m_firstReadDir(true), m_findDelayedPresent(false) {}
		~FsDir() { CloseFindHandle(); }

		FsDir* Dir() { return this; }
		void CloseFindHandle() { if (ValidFindHandle()) { FindClose(m_findHandle); m_findHandle = 0; } }
		bool ValidFindHandle() { return ValidHandle(m_findHandle); }
				
		HANDLE m_findHandle;
		bool m_noMoreFiles; // a file can still remain in m_nameDelayed
		bool m_firstReadDir;
		bool m_findDelayedPresent; // remaining from the previous listing
		WIN32_FIND_DATA m_findDelayed; 

		SFS::ListDir m_listDir;
	};

	class FsFile : public FsObj
	{
	public:
		FsFile();
		~FsFile() { CloseFileHandle(); }

		FsFile* File() { return this; }
		void CloseFileHandle() { if (ValidFileHandle()) { CloseHandle(m_fileHandle); m_fileHandle = 0; } }
		bool ValidFileHandle() { return ValidHandle(m_fileHandle); }

		HANDLE m_fileHandle;
		DWORD m_fileAccess; // desired acess flags
		
		bool m_append;
		bool m_textMode;
		unsigned __int64 m_textOffset;

		unsigned int m_refCount;

		SFS::TransferFile m_transferFile; // transfer statistics
	};

	// Takes ownership of the pased pointer on
	// success, otherwise an WException is thrown.
	void AddFsObj(FsObj* fsObj);
	// Throws WException if invalid SFS handle is provided.
	void DelFsObj(unsigned __int64 sfsHandle);

	unsigned __int64 m_lastSfsHandle;
	typedef std::map<unsigned __int64, FsObj*> FsObjMap;
	FsObjMap m_fsObjs;

private:
	struct WinTime
	{
		WinTime() : m_present(false) {}

		bool m_present;
		FILETIME m_time;
		
		FILETIME const* Get() const { return m_present ? &m_time : 0; }
	};

	struct WinTimes
	{
		WinTime m_createTime;
		WinTime m_modifyTime;
		WinTime m_accessTime;
		bool AnyPresent() const { return m_createTime.m_present || m_modifyTime.m_present || m_accessTime.m_present; }
	};

	// Returns false on failure.
	static bool GetSize(HANDLE handle, ULARGE_INTEGER& size);
	
	static void GetFileInfoT(HANDLE handle, BY_HANDLE_FILE_INFORMATION& fileInfo);
	static void GetFileInfoT(wchar_t const* winPath, BY_HANDLE_FILE_INFORMATION& fileInfo);
	static void GetFileInfoT(wchar_t const* winPath, WIN32_FILE_ATTRIBUTE_DATA& fileInfo);

	// Return false on failure. use GetLastError() for more info.
	static bool IsFile(WIN32_FIND_DATA const& findData);
	static bool IsFile(BY_HANDLE_FILE_INFORMATION const& fileInfo);
	static bool IsFile(wchar_t const* winPath);
	static bool IsDirectory(WIN32_FIND_DATA const& findData) { return !IsFile(findData); }
	static bool IsDirectory(BY_HANDLE_FILE_INFORMATION const& fileInfo) { return !IsFile(fileInfo); }
	static bool IsDirectory(wchar_t const* winPath);
	
	// Returns false for files/directories that will not appear in file listing.
	bool FindFileFilter(WIN32_FIND_DATA const& findData);
		
	// - Throws WinApiError on failure.
	// - ShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE.
	static HANDLE OpenExistingT(wchar_t const* winPath, DWORD access, DWORD flags);
	
	// - Throws WinApiError on failure.
	// - Handle must have FILE_DATA_WRITE | FILE_DATA_APPEND access.
	static void SetSizeT(HANDLE handle, unsigned __int64 size);
	// - Handle must have FILE_WRITE_ATTRIBUTES access.
	static void SetTimesT(HANDLE handle, WinTimes const& times);

	// - Throws WinApiError or failure, and returns true when reaching EOF.
	// - Handle must be in overlapped mode with FILE_READ_DATA access.
	bool ReadT(HANDLE handle, unsigned __int64 offset, byte* data, unsigned int& size);
	// - Handle must be in overlapped mode with FILE_DATA_WRITE | FILE_DATA_APPEND access.
	void WriteT(HANDLE handle, unsigned __int64 offset, byte const* data, unsigned int size);

	OVERLAPPED m_overlapped;
	
private:
	void SfsToWinPath(wchar_t const* sfsPath, std::wstring& winPath) const;

	// Throws WException if an invalid request is provided.
	wchar_t const* ExtractSfsPath(SFS_Request const* request);
	void ExtractWinPath(SFS_Request const* request, std::wstring& winPath);
	void ExtractWinPaths(SFS_Request const* request, std::wstring& winPath1, std::wstring& winPath2);

	// Throws WException if an invalid request is provided.
	unsigned __int64 ExtractSfsHandle(SFS_Request const* request);

	// Throws SfsStatusResponse if the handle is unrecognized.
	// Throws WException if an invalid request is provided.
	FsObj* ExtractFsObj(SFS_Request const* request);
	FsFile* ExtractFsFile(SFS_Request const* request);
	FsObj* ExtractFsRootOrDir(SFS_Request const* request);
	
	// Throws WException if an invalid request is provided.
	static SFS_Attrs const* ExtractSfsAttrs(SFS_Request const* request);
	void ExtractWinAttrs_PrepareLog(SFS_Request const* request, SFS::Log& log); // Adds 2-3 log items 
	wchar_t const* ExtractWinAttrs_GetWhileSuffix(SFS_Request const* request);
	
	// Throws SfsStatusResponse on SfsToWinTime() failure.
	// Throws WException if an invalid request is provided.
	void ExtractWinTimes(SFS_Request const* request, WinTimes& winTimes);
	
	void ToSfsAttrs(WIN32_FIND_DATA const& data, SFS_Attrs& sfsAttrs);
	void ToSfsAttrs(WIN32_FILE_ATTRIBUTE_DATA const& data, SFS_Attrs& sfsAttrs);
	void ToSfsAttrs(BY_HANDLE_FILE_INFORMATION const& info, SFS_Attrs& sfsAttrs);
	
	// See SetFileAttributes() for more info.
	static DWORD RemoveReadOnlyWinAttrBits(DWORD winAttrs);
	// Basic bits = ReadOnly, System, Hidden, Archive
	static DWORD ModifyBasicWinAttrBits(DWORD winAttrs, SFS_Attrs const* sfsAttrs);

	static bool ToWinTime(unsigned __int64 sfsTime, unsigned int sfsTimeNs, FILETIME& winTime);
	static bool ToSfsTime(FILETIME const& winTime, unsigned __int64& sfsTime, unsigned int& sfsTimeNs);

	DWORD ToWinAccessRights(unsigned int aceMask);

	static unsigned int ToSfsStatusCode(DWORD winError);
	static std::wstring ToStringRequestType(unsigned int requestType);
	
private:
	std::wstring m_mountPath;

	static unsigned int const s_maxReadSize;
	
	enum LogStatus
	{
		EnablePrivilegeFailure = 201,
		ConvertTimeFailure,
		ConvertSidFailure,
		CreateAclFailure,
		CreateSdFailure,

		SetFileTimeFailureAsProgress = 601,
		RetrieveTextHintFailureAsProgress,
		SetFileAttrsFailureAsProgress,
		GetFileTypeFailureAsProgress,
	};
};
