#pragma once

// WinAPI
#include <windows.h>

// STL
#include <string>


class NoCopy
{
private:
	NoCopy(NoCopy const& x);
	void operator= (NoCopy const& x);
public:
	NoCopy() {}
};


template <class T>
class AutoArray : public NoCopy
{
public:
	AutoArray(T* p = 0) : m_p(p) {}
	~AutoArray() { Reset(); }

	void Reset() { Set(0); }
	void Set(T* p)
	{
		if (m_p)
			delete [] m_p;
		m_p = p;
	}
	
	T* Get() const { return m_p; }
	operator T*() const { return m_p; }

private:
	T* m_p;
};


class AutoHandle : public NoCopy
{
public:
	AutoHandle(HANDLE h = 0) : m_h(h) {}
	~AutoHandle() { Reset(); }

	bool Valid() const { return m_h != 0 && m_h != INVALID_HANDLE_VALUE; }
	void Reset() { Set(0); }
	void Set(HANDLE h);
	
	HANDLE Get() const { return m_h; }
	operator HANDLE() const { return m_h; }

private:
	HANDLE m_h;
};


class AutoCryptContext : public NoCopy
{
public:
	AutoCryptContext() : m_valid(false) {}
	AutoCryptContext(HCRYPTPROV prov)  : m_valid(true), m_prov(prov) {}	
	~AutoCryptContext() { Reset(); }

	bool Valid() const { return m_valid; }
	void Reset();
	void Set(HCRYPTPROV prov);

	HCRYPTPROV Get() const { return m_valid ? m_prov : 0; }
	operator HCRYPTPROV() const { return Get(); }

private:
	bool m_valid;
	HCRYPTPROV m_prov;
};

class AutoCryptHash : public NoCopy
{
public:
	AutoCryptHash() : m_valid(false) {}
	AutoCryptHash(HCRYPTHASH hash) : m_valid(true), m_hash(hash) {}	
	~AutoCryptHash() { Reset(); }

	bool Valid() const { return m_valid; }
	void Reset();
	void Set(HCRYPTHASH hash);

	HCRYPTHASH Get() const { return m_valid ? m_hash : 0; }
	operator HCRYPTHASH() const { return Get(); }

private:
	bool m_valid;
	HCRYPTHASH m_hash;
};


class WException : virtual public std::exception
{
public:
	virtual wchar_t const* What() const = 0;
	char const* what() const;

private:
	mutable std::string m_a;
};

class WStrException : virtual public WException
{
public:
	WStrException(wchar_t const* s) : m_s(s) {}
	WStrException(std::wstring const& s) : m_s(s) {}

	wchar_t const* What() const { return m_s.c_str(); }

private:
	std::wstring m_s;
};

class WinApiError : virtual public WException
{
public:
	WinApiError(wchar_t const* operation, DWORD errorCode = GetLastError())
		: m_operation(operation), m_errorCode(errorCode) {}

	wchar_t const* What() const;
	wchar_t const* GetOperation() const { return m_operation.c_str(); }
	DWORD GetErrorCode() const { return m_errorCode; }
	
private:
	DWORD m_errorCode;
	std::wstring m_operation;
	mutable std::wstring m_what;
};


bool WideToNarrow(wchar_t const* wide, std::string& narrow, unsigned int codePage);
bool NarrowToWide(char const* narrow, unsigned int codePage, std::wstring& wide);


bool UxtimeToFtime(time_t uxt, FILETIME& ft);
bool FtimeToUxtime(FILETIME ft, time_t& uxt);
